# Copyright (C) 2018-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict

from attr import asdict, has
from confluent_kafka import Consumer
from hypothesis import given
from hypothesis.strategies import lists

from swh.journal.pytest_plugin import assert_all_objects_consumed, consume_messages
from swh.model.hypothesis_strategies import objects
from swh.model.model import ModelObjectType, Person
from swh.model.tests.swh_model_data import TEST_OBJECTS
from swh.storage import get_storage


def test_storage_direct_writer(kafka_prefix: str, kafka_server, consumer: Consumer):
    writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "anonymize": False,
        "auto_flush": False,
    }
    storage_config: Dict[str, Any] = {
        "cls": "pipeline",
        "steps": [
            {"cls": "memory", "journal_writer": writer_config},
        ],
    }

    object_types = (
        ModelObjectType.CONTENT,
        ModelObjectType.SKIPPED_CONTENT,
        ModelObjectType.DIRECTORY,
        ModelObjectType.EXTID,
        ModelObjectType.METADATA_AUTHORITY,
        ModelObjectType.METADATA_FETCHER,
        ModelObjectType.REVISION,
        ModelObjectType.RELEASE,
        ModelObjectType.SNAPSHOT,
        ModelObjectType.ORIGIN,
        ModelObjectType.ORIGIN_VISIT,
        ModelObjectType.ORIGIN_VISIT_STATUS,
        ModelObjectType.RAW_EXTRINSIC_METADATA,
    )

    storage = get_storage(**storage_config)

    expected_messages = 0

    for obj_type, objs in TEST_OBJECTS.items():
        method = getattr(storage, f"{obj_type}_add")
        if obj_type in object_types:
            method(objs)
            expected_messages += len(objs)
        else:
            assert False, obj_type
    storage.journal_writer.journal.flush()  # type: ignore[attr-defined]

    existing_topics = set(
        topic
        for topic in consumer.list_topics(timeout=10).topics.keys()
        if topic.startswith(f"{kafka_prefix}.")  # final . to exclude privileged topics
    )
    assert existing_topics == {
        f"{kafka_prefix}.{obj_type}" for obj_type in object_types
    }

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    assert_all_objects_consumed(consumed_messages)


def test_storage_direct_writer_anonymized(
    kafka_prefix: str, kafka_server, consumer: Consumer
):
    writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "anonymize": True,
        "auto_flush": False,
    }
    storage_config: Dict[str, Any] = {
        "cls": "pipeline",
        "steps": [
            {"cls": "memory", "journal_writer": writer_config},
        ],
    }

    storage = get_storage(**storage_config)

    expected_messages = 0

    for obj_type, objs in TEST_OBJECTS.items():
        if obj_type == ModelObjectType.ORIGIN_VISIT:
            # these have non-consistent API and are unrelated with what we
            # want to test here
            continue
        method = getattr(storage, f"{obj_type}_add")
        method(objs)
        expected_messages += len(objs)
    storage.journal_writer.journal.flush()  # type: ignore[attr-defined]

    existing_topics = set(
        topic
        for topic in consumer.list_topics(timeout=10).topics.keys()
        if topic.startswith(kafka_prefix)
    )
    assert existing_topics == {
        f"{kafka_prefix}.{obj_type}"
        for obj_type in (
            "content",
            "directory",
            "extid",
            "metadata_authority",
            "metadata_fetcher",
            "origin",
            "origin_visit",
            "origin_visit_status",
            "raw_extrinsic_metadata",
            "release",
            "revision",
            "snapshot",
            "skipped_content",
        )
    } | {
        f"{kafka_prefix}_privileged.{obj_type}"
        for obj_type in (
            "release",
            "revision",
        )
    }


def check_anonymized_obj(obj):
    if has(obj):
        if isinstance(obj, Person):
            assert obj.name is None
            assert obj.email is None
            assert len(obj.fullname) == 32
        else:
            for key, value in asdict(obj, recurse=False).items():
                check_anonymized_obj(value)


@given(lists(objects(split_content=True)))
def test_anonymizer(obj_type_and_objs):
    for obj_type, obj in obj_type_and_objs:
        check_anonymized_obj(obj.anonymize())



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from unittest.mock import patch

from swh.model.from_disk import DentryPerms
from swh.model.hashutil import MultiHash, hash_to_bytes
from swh.model.model import Directory
from swh.storage.algos.dir_iterators import dir_iterator

# flake8: noqa


class DirectoryModel(object):
    """
    Quick and dirty directory model to ease the writing
    of directory iterators and revision trees differential tests.
    """

    def __init__(self, name=""):
        self.data = {}
        self.data["name"] = name
        self.data["perms"] = DentryPerms.directory
        self.data["type"] = "dir"
        self.data["entries"] = []
        self.data["entry_idx"] = {}

    def __getitem__(self, item):
        if item == "target":
            return Directory.from_dict(
                {
                    "entries": [
                        {
                            "name": entry["name"],
                            "target": entry["target"],
                            "type": entry["type"],
                            "perms": entry["perms"],
                        }
                        for entry in self.data["entries"]
                    ]
                }
            ).id
        else:
            return self.data[item]

    def add_file(self, path, sha1=None):
        path_parts = path.split(b"/")
        sha1 = (
            hash_to_bytes(sha1) if sha1 else MultiHash.from_data(path).digest()["sha1"]
        )
        if len(path_parts) == 1:
            self["entry_idx"][path] = len(self["entries"])
            self["entries"].append(
                {
                    "target": sha1,
                    "name": path,
                    "perms": DentryPerms.content,
                    "type": "file",
                }
            )
        else:
            if not path_parts[0] in self["entry_idx"]:
                self["entry_idx"][path_parts[0]] = len(self["entries"])
                self["entries"].append(DirectoryModel(path_parts[0]))
            if path_parts[1]:
                dir_idx = self["entry_idx"][path_parts[0]]
                self["entries"][dir_idx].add_file(b"/".join(path_parts[1:]), sha1)

    def get_hash_data(self, entry_hash):
        if self["target"] == entry_hash:
            ret = []
            for e in self["entries"]:
                ret.append(
                    {
                        "target": e["target"],
                        "name": e["name"],
                        "perms": e["perms"],
                        "type": e["type"],
                    }
                )
            return ret
        else:
            for e in self["entries"]:
                if e["type"] == "file" and e["target"] == entry_hash:
                    return e
                elif e["type"] == "dir":
                    data = e.get_hash_data(entry_hash)
                    if data:
                        return data
            return None

    def get_path_data(self, path):
        path_parts = path.split(b"/")
        entry_idx = self["entry_idx"][path_parts[0]]
        entry = self["entries"][entry_idx]
        if len(path_parts) == 1:
            return {
                "target": entry["target"],
                "name": entry["name"],
                "perms": entry["perms"],
                "type": entry["type"],
            }
        else:
            return entry.get_path_data(b"/".join(path_parts[1:]))


@patch("swh.storage.algos.dir_iterators._get_dir")
class TestDirectoryIterator(unittest.TestCase):
    def check_iterated_paths(self, dir_model, expected_paths_order, mock_get_dir):
        def _get_dir(*args, **kwargs):
            return dir_model.get_hash_data(args[1])

        mock_get_dir.side_effect = _get_dir  # noqa
        paths_order = [e["path"] for e in dir_iterator(None, dir_model["target"])]
        self.assertEqual(paths_order, expected_paths_order)

    def test_dir_iterator_empty_dir(self, mock_get_dir):
        dir_model = DirectoryModel()
        expected_paths_order = []
        self.check_iterated_paths(dir_model, expected_paths_order, mock_get_dir)

    def test_dir_iterator_no_empty_dirs(self, mock_get_dir):
        dir_model = DirectoryModel()
        dir_model.add_file(b"xyz/gtr/uhb")
        dir_model.add_file(b"bca/ef")
        dir_model.add_file(b"abc/ab")
        dir_model.add_file(b"abc/bc")
        dir_model.add_file(b"xyz/ouy/poi")

        expected_paths_order = [
            b"abc",
            b"abc/ab",
            b"abc/bc",
            b"bca",
            b"bca/ef",
            b"xyz",
            b"xyz/gtr",
            b"xyz/gtr/uhb",
            b"xyz/ouy",
            b"xyz/ouy/poi",
        ]

        self.check_iterated_paths(dir_model, expected_paths_order, mock_get_dir)

    def test_dir_iterator_with_empty_dirs(self, mock_get_dir):
        dir_model = DirectoryModel()
        dir_model.add_file(b"xyz/gtr/")
        dir_model.add_file(b"bca/ef")
        dir_model.add_file(b"abc/")
        dir_model.add_file(b"xyz/ouy/poi")

        expected_paths_order = [
            b"abc",
            b"bca",
            b"bca/ef",
            b"xyz",
            b"xyz/gtr",
            b"xyz/ouy",
            b"xyz/ouy/poi",
        ]

        self.check_iterated_paths(dir_model, expected_paths_order, mock_get_dir)



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import closing
import datetime as dt
from io import StringIO
import socket
import textwrap

from click.testing import CliRunner
import pytest

from swh.core.cli.db import db as swhdb
from swh.core.db.db_utils import get_database_info
from swh.core.tests.test_cli import assert_result
from swh.storage.tests.storage_data import StorageData as data

from ...proxies.blocking.cli import (
    EditAborted,
    blocking_cli_group,
    clear_request,
    edit_message,
    history_cmd,
    list_requests,
    new_request,
    object_state,
    read_origins,
    status_cmd,
    update_objects,
)
from ...proxies.blocking.db import (
    BlockedOrigin,
    BlockingAdmin,
    BlockingRequest,
    BlockingState,
    RequestHistory,
    RequestNotFound,
)


@pytest.mark.parametrize(
    "module_name", ["storage:blocking", "storage.proxies.blocking"]
)
def test_cli_db_create(postgresql, module_name):
    """Create a db then initializing it should be ok"""
    db_params = postgresql.info
    dbname = f"{module_name}-db"
    conninfo = (
        f"postgresql://{db_params.user}@{db_params.host}:{db_params.port}/{dbname}"
    )

    # This creates the db and installs the necessary admin extensions
    result = CliRunner().invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert_result(result)

    # This initializes the schema and data
    result = CliRunner().invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    assert_result(result)

    dbmodule, dbversion, dbflavor = get_database_info(conninfo)
    assert dbmodule == "storage:blocking"
    assert dbversion == BlockingAdmin.current_version
    assert dbflavor is None


@pytest.fixture
def blocking_admin_config(blocking_db_postgresql):
    return {
        "blocking_admin": {"cls": "postgresql", "db": blocking_db_postgresql.info.dsn}
    }


def test_blocking_admin_not_defined():
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        blocking_cli_group,
        ["list-requests"],
        obj={"config": {}},
    )
    assert result.exit_code == 2
    assert "blocking_admin" in result.stderr


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("localhost", 0))
        # We explicitly not set SO_REUSEADDR because we want the port
        # to stay free so we can get our connection refused.
        return s.getsockname()[1]


def test_blocking_admin_unreachable():
    erroneous_postgresql_port = find_free_port()
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        blocking_cli_group,
        ["list-requests"],
        obj={
            "config": {
                "blocking_admin": {
                    "blocking_db": (
                        "postgresql://localhost/postgres?"
                        f"port={erroneous_postgresql_port}"
                    )
                }
            }
        },
    )
    assert result.exit_code == 1
    assert "failed: Connection refused" in result.stderr


def test_edit_message_success(mocker):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.blocking.cli.click.edit",
        return_value="Sometimes I'm round",
    )
    assert (
        edit_message("Hello!", extra_lines=["Sometimes I'm alone", "Sometimes I'm not"])
        == "Sometimes I'm round"
    )
    mocked_click_edit.assert_called_once_with(
        textwrap.dedent(
            """

        # Hello!
        # Lines starting with “#” will be ignored. An empty message will abort the operation.
        #
        # Sometimes I'm alone
        # Sometimes I'm not"""
        )
    )


def test_edit_message_removes_comment(mocker):
    mocker.patch(
        "swh.storage.proxies.blocking.cli.click.edit",
        return_value=textwrap.dedent(
            """\
        Hello!

        # This is a comment"""
        ),
    )
    assert edit_message("RANDOM PROMPT") == "Hello!"


def test_edit_message_empty_message_aborts(mocker):
    mocker.patch("swh.storage.proxies.blocking.cli.click.edit", return_value="")
    with pytest.raises(EditAborted):
        edit_message("RANDOM PROMPT")


@pytest.fixture
def request01():
    return BlockingRequest(
        id="da785a27-7e59-4a35-b82a-a5ae3714407c",
        slug="request-01",
        date=dt.datetime.now(dt.timezone.utc),
        reason="one",
    )


@pytest.fixture
def request02():
    return BlockingRequest(
        id="4fd42e35-2b6c-4536-8447-bc213cd0118b",
        slug="request-02",
        date=dt.datetime.now(dt.timezone.utc),
        reason="two",
    )


@pytest.fixture
def blocked_origin(request01):
    return BlockedOrigin(
        request_slug=request01.slug,
        url_pattern=data.origins[0].url,
        state=BlockingState.BLOCKED,
    )


@pytest.fixture
def blocked_origin2(request01):
    return BlockedOrigin(
        request_slug=request01.slug,
        url_pattern=data.origins[1].url,
        state=BlockingState.DECISION_PENDING,
    )


def blocked_origin3(request02):
    return BlockedOrigin(
        request_slug=request02.slug,
        url_pattern=data.origins[2].url,
        state=BlockingState.NOT_BLOCKED,
    )


@pytest.fixture
def history01(request01):
    return RequestHistory(
        request=request01.id,
        date=dt.datetime.now(dt.timezone.utc),
        message="record one",
    )


@pytest.fixture
def history02(request01):
    return RequestHistory(
        request=request01.id,
        date=dt.datetime.now(dt.timezone.utc),
        message="record two",
    )


@pytest.fixture
def mocked_blocking_admin(
    mocker, request01, request02, blocked_origin, blocked_origin2, history01, history02
):
    def mock_find_request(slug):
        if slug == "request-01":
            return request01
        elif slug == "request-02":
            return request02
        else:
            return None

    def mock_find_request_by_id(id):
        if str(id) == "da785a27-7e59-4a35-b82a-a5ae3714407c":
            return request01
        return None

    def mock_get_states_for_request(request_id):
        if request_id == request01.id:
            return {
                blocked_origin.url_pattern: blocked_origin.state,
                blocked_origin2.url_pattern: blocked_origin2.state,
            }
        return RequestNotFound(request_id)

    def mock_get_history(request_id):
        if request_id == request01.id:
            return [history02, history01]
        return RequestNotFound(request_id)

    mock = mocker.Mock(spec=BlockingAdmin)
    mock.find_request.side_effect = mock_find_request
    mock.find_request_by_id.side_effect = mock_find_request_by_id
    mock.get_states_for_request.side_effect = mock_get_states_for_request
    mock.get_history.side_effect = mock_get_history
    mock.get_requests.return_value = [(request02, 2), (request01, 1)]
    return mock


def test_new_request(mocker, mocked_blocking_admin, request01):
    mocked_blocking_admin.create_request.return_value = request01
    mocker.patch("swh.storage.proxies.blocking.cli.click.edit", return_value="one")
    runner = CliRunner()
    result = runner.invoke(
        new_request, ["request-01"], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert_result(result)
    mocked_blocking_admin.create_request.assert_called_once_with("request-01", "one")
    assert "da785a27-7e59-4a35-b82a-a5ae3714407c" in result.output
    assert "request-01" in result.output


def test_new_request_with_message(mocker, mocked_blocking_admin, request01):
    mocked_blocking_admin.create_request.return_value = request01
    mocked_click_edit = mocker.patch("swh.storage.proxies.blocking.cli.click.edit")
    runner = CliRunner()
    result = runner.invoke(
        new_request,
        ["--message=one", "request-01"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    mocked_click_edit.assert_not_called()
    mocked_blocking_admin.create_request.assert_called_once_with("request-01", "one")
    assert "da785a27-7e59-4a35-b82a-a5ae3714407c" in result.output
    assert "request-01" in result.output


def test_new_request_empty_message_aborts(mocker, mocked_blocking_admin):
    mocked_blocking_admin.create_request.return_value = request01
    mocker.patch("swh.storage.proxies.blocking.cli.click.edit", side_effect=EditAborted)
    runner = CliRunner()
    result = runner.invoke(
        new_request,
        ["request-01"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert result.exit_code == 1
    assert "Aborting" in result.output
    mocked_blocking_admin.create_request.assert_not_called()


def test_list_requests_default(mocked_blocking_admin):
    runner = CliRunner()
    result = runner.invoke(
        list_requests, [], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert_result(result)
    mocked_blocking_admin.get_requests.assert_called_once_with(
        include_cleared_requests=False
    )
    assert "request-01" in result.output
    assert "da785a27-7e59-4a35-b82a-a5ae3714407c" in result.output
    assert "request-02" in result.output
    assert "4fd42e35-2b6c-4536-8447-bc213cd0118b" in result.output


def test_list_requests_include_cleared_requests(mocked_blocking_admin):
    runner = CliRunner()
    result = runner.invoke(
        list_requests,
        ["--include-cleared-requests"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    mocked_blocking_admin.get_requests.assert_called_once_with(
        include_cleared_requests=True
    )


ORIGINS_INPUT = f"""\
# origin1
{data.origin.url}

# origin2
{data.origin2.url}
"""

ORIGINS_INPUT_PARSED = [
    data.origin.url,
    data.origin2.url,
]


def test_read_origins():
    file = StringIO(ORIGINS_INPUT)
    assert read_origins(file) == ORIGINS_INPUT_PARSED


def test_update_objects(mocker, mocked_blocking_admin, request01):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.blocking.cli.click.edit", return_value="action made"
    )
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending"],
        input=ORIGINS_INPUT,
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    assert "2 origins" in result.output
    assert "request-01" in result.output
    mocked_click_edit.assert_called_once()
    mocked_blocking_admin.set_origins_state.assert_called_once_with(
        request01.id, BlockingState.DECISION_PENDING, ORIGINS_INPUT_PARSED
    )
    mocked_blocking_admin.record_history.assert_called_once_with(
        request01.id, "action made"
    )


def test_update_objects_with_message(mocker, mocked_blocking_admin, request01):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.blocking.cli.click.edit", return_value="action made"
    )
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", "--message=action made"],
        input=ORIGINS_INPUT,
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    mocked_click_edit.assert_not_called()
    mocked_blocking_admin.set_origins_state.assert_called_once()
    mocked_blocking_admin.record_history.assert_called_once_with(
        request01.id, "action made"
    )


def test_update_objects_empty_message_aborts(mocker, mocked_blocking_admin):
    mocker.patch("swh.storage.proxies.blocking.cli.click.edit", side_effect=EditAborted)
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending"],
        input=ORIGINS_INPUT,
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert result.exit_code == 1
    mocked_blocking_admin.record_history.assert_not_called()


def test_update_objects_with_file(tmp_path, mocker, mocked_blocking_admin, request01):
    mocker.patch(
        "swh.storage.proxies.blocking.cli.click.edit", return_value="action made"
    )
    origins_file = tmp_path / "origins"
    origins_file.write_text(ORIGINS_INPUT)
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", f"--file={str(origins_file)}"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    assert "2 origins" in result.output
    mocked_blocking_admin.set_origins_state.assert_called_once_with(
        request01.id, BlockingState.DECISION_PENDING, ORIGINS_INPUT_PARSED
    )


def test_update_objects_no_origins(mocked_blocking_admin):
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", "--message=message"],
        input="",
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert result.exit_code == 1
    assert "No origin given!" in result.output
    mocked_blocking_admin.set_origins_state.assert_not_called()
    mocked_blocking_admin.record_history.assert_not_called()


def test_status(mocked_blocking_admin, request01):
    runner = CliRunner()
    result = runner.invoke(
        status_cmd, ["request-01"], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert_result(result)
    mocked_blocking_admin.get_states_for_request.assert_called_once_with(request01.id)
    assert result.output == textwrap.dedent(
        """\
        https://github.com/user1/repo1 blocked
        https://github.com/user2/repo1 decision-pending
        """
    )


def test_status_request_not_found(mocked_blocking_admin):
    runner = CliRunner()
    result = runner.invoke(
        status_cmd, ["garbage"], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert result.exit_code == 1
    assert "Error: Request “garbage” not found" in result.output


def test_status_request_via_uuid(mocked_blocking_admin, request01):
    runner = CliRunner()
    result = runner.invoke(
        status_cmd,
        ["da785a27-7e59-4a35-b82a-a5ae3714407c"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    mocked_blocking_admin.get_states_for_request.assert_called_once_with(request01.id)


def test_history(mocked_blocking_admin, request01):
    runner = CliRunner()
    result = runner.invoke(
        history_cmd, ["request-01"], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert_result(result)
    mocked_blocking_admin.get_history.assert_called_once_with(request01.id)
    assert (
        "History for request “request-01” (da785a27-7e59-4a35-b82a-a5ae3714407c)"
        in result.output
    )
    assert "record two" in result.output
    assert "record one" in result.output


def test_history_not_found(mocked_blocking_admin):
    runner = CliRunner()
    result = runner.invoke(
        history_cmd, ["garbage"], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert result.exit_code == 1
    assert "Error: Request “garbage” not found" in result.output


def test_object_state(mocked_blocking_admin, blocked_origin, blocked_origin2):
    blocked_origin_in_request02 = BlockedOrigin(
        request_slug="request-02",
        url_pattern=blocked_origin.url_pattern,
        state=BlockingState.NON_BLOCKED,
    )
    mocked_blocking_admin.find_blocking_states.return_value = [
        blocked_origin_in_request02,
        blocked_origin,
        blocked_origin2,
    ]
    runner = CliRunner()
    result = runner.invoke(
        object_state,
        [
            data.origins[0].url,
            data.origins[1].url,
            data.origins[2].url,
        ],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    assert result.output == textwrap.dedent(
        """\
        blocked  https://github.com/user1/repo1
                request-02: allowed
                request-01: blocked
        blocked  https://github.com/user2/repo1
                request-01: decision-pending

        """
    )


def test_clear_request(mocker, mocked_blocking_admin, request01):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.blocking.cli.click.edit", return_value="request cleared"
    )
    runner = CliRunner()
    result = runner.invoke(
        clear_request,
        ["request-01"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    assert "Blockings cleared for request “request-01”." in result.output
    mocked_click_edit.assert_called_once()
    mocked_blocking_admin.delete_blocking_states.assert_called_once_with(request01.id)
    mocked_blocking_admin.record_history.assert_called_once_with(
        request01.id, "request cleared"
    )


def test_clear_request_with_message(mocker, mocked_blocking_admin, request01):
    mocked_click_edit = mocker.patch("swh.storage.proxies.blocking.cli.click.edit")
    runner = CliRunner()
    result = runner.invoke(
        clear_request,
        ["--message=request cleared", "request-01"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert_result(result)
    assert "Blockings cleared for request “request-01”." in result.output
    mocked_click_edit.assert_not_called()
    mocked_blocking_admin.delete_blocking_states.assert_called_once_with(request01.id)
    mocked_blocking_admin.record_history.assert_called_once_with(
        request01.id, "request cleared"
    )


def test_clear_request_empty_message_aborts(mocker, mocked_blocking_admin):
    mocker.patch("swh.storage.proxies.blocking.cli.click.edit", side_effect=EditAborted)
    runner = CliRunner()
    result = runner.invoke(
        clear_request,
        ["request-01"],
        obj={"blocking_admin": mocked_blocking_admin},
    )
    assert result.exit_code == 1
    assert "Aborting" in result.output
    mocked_blocking_admin.delete_blocking_states.assert_not_called()
    mocked_blocking_admin.record_history.assert_not_called()


def test_clear_request_not_found(mocked_blocking_admin):
    runner = CliRunner()
    result = runner.invoke(
        clear_request, ["garbage"], obj={"blocking_admin": mocked_blocking_admin}
    )
    assert result.exit_code == 1
    assert "Error: Request “garbage” not found" in result.output
    mocked_blocking_admin.delete_blocking_states.assert_not_called()
    mocked_blocking_admin.record_history.assert_not_called()



# Copyright (C) 2018-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dataclasses
import datetime
import itertools
import re
from typing import Any, Dict, Union

import attr
from cassandra.cluster import NoHostAvailable
import pytest

from swh.core.api.classes import stream_results
from swh.model import from_disk
from swh.model.model import (
    Directory,
    DirectoryEntry,
    ExtID,
    Person,
    Snapshot,
    SnapshotBranch,
    TimestampWithTimezone,
)
from swh.model.swhids import CoreSWHID, ObjectType
from swh.storage import get_storage
from swh.storage.cassandra.cql import BATCH_INSERT_MAX_SIZE
import swh.storage.cassandra.model
from swh.storage.cassandra.model import BaseRow, ContentRow, ExtIDRow
from swh.storage.cassandra.schema import CREATE_TABLES_QUERIES, HASH_ALGORITHMS, TABLES
from swh.storage.cassandra.storage import (
    DIRECTORY_ENTRIES_INSERT_ALGOS,
    CassandraStorage,
)
from swh.storage.tests.storage_data import StorageData
from swh.storage.tests.storage_tests import (
    TestStorageGeneratedData as _TestStorageGeneratedData,
)
from swh.storage.tests.storage_tests import TestStorage as _TestStorage
from swh.storage.utils import now, remove_keys


@pytest.fixture
def swh_storage_backend_config(swh_storage_cassandra_backend_config):
    return swh_storage_cassandra_backend_config


def _python_type_to_cql_type_re(ty: type) -> str:
    if ty is bytes:
        return "blob"
    elif ty is str:
        return "(ascii|text)"
    elif ty is int:
        return "(small||big)int"
    elif ty is datetime.datetime:
        return "timestamp"
    elif ty is TimestampWithTimezone:
        return "microtimestamp_with_timezone"
    elif ty is Person:
        return "person"
    elif ty is bool:
        return "boolean"
    elif ty is dict:
        return "frozen<list <list<blob>> >"
    elif getattr(ty, "__origin__", None) is Union:
        if len(ty.__args__) == 2 and type(None) in ty.__args__:  # type: ignore
            (inner_type,) = [
                arg
                for arg in ty.__args__  # type: ignore
                if arg is not type(None)  # noqa
            ]
            # This is Optional[inner_type]. CQL does not support NOT NULL,
            # so the corresponding type is:
            return _python_type_to_cql_type_re(inner_type)

    assert False, f"Unsupported type: {ty}"


@pytest.mark.cassandra
def test_schema_matches_model():
    """Checks tables defined in :mod:`swh.storage.cassandra.schema` match
    the object model defined in :mod:`swh.storage.cassandra.model`.
    """

    models = {}  # {table_name: cls}
    for obj_name in dir(swh.storage.cassandra.model):
        cls = getattr(swh.storage.cassandra.model, obj_name)
        if isinstance(cls, type) and issubclass(cls, BaseRow) and hasattr(cls, "TABLE"):
            assert cls.TABLE not in models, f"Duplicate model for table {cls.TABLE}"
            models[cls.TABLE] = cls

    del models["object_count"]  # https://forge.softwareheritage.org/D6150

    statements = {}  # {table_name: statement}
    for statement in CREATE_TABLES_QUERIES:
        statement = re.sub(" +", " ", statement.strip())  # normalize whitespace
        if statement.startswith(
            (
                "CREATE TYPE ",
                "CREATE OR REPLACE FUNCTION",
                "CREATE OR REPLACE AGGREGATE",
                "-- Secondary table",
            )
        ):
            continue
        prefix = "CREATE TABLE IF NOT EXISTS "
        assert statement.startswith(prefix)
        table_name = statement[len(prefix) :].split(" ", 1)[0]
        assert (
            table_name not in statements
        ), f"Duplicate statement for table {table_name}"

        statements[table_name] = statement

    assert set(models) - set(statements) == set(), "Missing statements"
    assert set(statements) - set(models) == set(), "Missing models"

    mismatches = []
    for table_name in list(models):
        # Compute what we expect the statement to look like:
        model = models[table_name]
        expected_lines = [rf"CREATE TABLE IF NOT EXISTS {model.TABLE} \("]
        for field in dataclasses.fields(model):
            expected_lines.append(
                f" {field.name} {_python_type_to_cql_type_re(field.type)},"
            )

        partition_key = ", ".join(model.PARTITION_KEY)
        clustering_key = "".join(", " + col for col in model.CLUSTERING_KEY)
        expected_lines.append(rf" PRIMARY KEY \(\({partition_key}\){clustering_key}\)")

        if table_name == "origin_visit_status":
            # we need to special-case this one
            expected_lines.append(r"\)")
            expected_lines.append(r"WITH CLUSTERING ORDER BY \(visit DESC, date DESC\)")
            expected_lines.append(r";")
        else:
            expected_lines.append(r"\);")

        statement = statements[table_name]
        actual_lines = statement.split("\n")

        mismatches.extend(
            (table_name, expected_line, actual_line)
            for (expected_line, actual_line) in zip(expected_lines, actual_lines)
            if not re.match(expected_line, actual_line)
        )

    assert not mismatches, "\n" + "\n".join(
        f"{table_name}:\t{actual_line!r} did not match {expected_line!r}"
        for (table_name, expected_line, actual_line) in mismatches
    )


# tests are executed using imported classes (TestStorage and
# TestStorageGeneratedData) using overloaded swh_storage fixture
# below


@pytest.mark.cassandra
class TestCassandraStorage(_TestStorage):
    def test_config_wrong_consistency_should_raise(self):
        storage_config = dict(
            cls="cassandra",
            hosts=["first"],
            port=9999,
            keyspace="any",
            consistency_level="fake",
            journal_writer={"cls": "memory"},
            objstorage={"cls": "memory"},
        )

        with pytest.raises(ValueError, match="Unknown consistency"):
            get_storage(**storage_config)

    def test_config_consistency_used(self, swh_storage_backend_config):
        config_with_consistency = dict(
            swh_storage_backend_config, **{"consistency_level": "THREE"}
        )

        storage = get_storage(**config_with_consistency)

        with pytest.raises(NoHostAvailable):
            storage.content_get_random()

    def test_no_authentication(self, swh_storage_backend_config):
        config_without_authentication = swh_storage_backend_config.copy()
        config_without_authentication.pop("auth_provider")

        with pytest.raises(NoHostAvailable):
            get_storage(**config_without_authentication)

    def test_no_authenticator_class(self, swh_storage_backend_config):
        config_without_authentication = swh_storage_backend_config.copy()
        authenticator_config = config_without_authentication.pop("auth_provider").copy()
        authenticator_config.pop("cls")
        config_without_authentication["auth_provider"] = authenticator_config

        with pytest.raises(ValueError, match="cls property"):
            get_storage(**config_without_authentication)

    def test_content_add_murmur3_collision(self, swh_storage, mocker, sample_data):
        """The Murmur3 token is used as link from index tables to the main
        table; and non-matching contents with colliding murmur3-hash
        are filtered-out when reading the main table.
        This test checks the content methods do filter out these collision.
        """
        called = 0

        cont, cont2 = sample_data.contents[:2]

        # always return a token
        def mock_cgtfsa(algo, hashes):
            nonlocal called
            called += 1
            assert algo in ("sha1", "sha1_git")
            return [123456]

        mocker.patch.object(
            swh_storage._cql_runner,
            "content_get_tokens_from_single_algo",
            mock_cgtfsa,
        )

        # For all tokens, always return cont
        def mock_cgft(tokens):
            nonlocal called
            called += 1
            return [
                ContentRow(
                    length=10,
                    ctime=datetime.datetime.now(),
                    status="present",
                    **{algo: getattr(cont, algo) for algo in HASH_ALGORITHMS},
                )
            ]

        mocker.patch.object(
            swh_storage._cql_runner, "content_get_from_tokens", mock_cgft
        )

        actual_result = swh_storage.content_add([cont2])

        assert called == 4
        assert actual_result == {
            "content:add": 1,
            "content:add:bytes": cont2.length,
        }

    def test_content_get_metadata_murmur3_collision(
        self, swh_storage, mocker, sample_data
    ):
        """The Murmur3 token is used as link from index tables to the main
        table; and non-matching contents with colliding murmur3-hash
        are filtered-out when reading the main table.
        This test checks the content methods do filter out these collisions.
        """
        called = 0

        cont, cont2 = [attr.evolve(c, ctime=now()) for c in sample_data.contents[:2]]

        # always return a token
        def mock_cgtfsa(algo, hashes):
            nonlocal called
            called += 1
            assert algo in ("sha1", "sha1_git")
            return [123456]

        mocker.patch.object(
            swh_storage._cql_runner,
            "content_get_tokens_from_single_algo",
            mock_cgtfsa,
        )

        # For all tokens, always return cont and cont2
        cols = list(set(cont.to_dict()) - {"data"})

        def mock_cgft(tokens):
            nonlocal called
            called += 1
            return [
                ContentRow(
                    **{col: getattr(cont, col) for col in cols},
                )
                for cont in [cont, cont2]
            ]

        mocker.patch.object(
            swh_storage._cql_runner, "content_get_from_tokens", mock_cgft
        )

        actual_result = swh_storage.content_get([cont.sha1])
        assert called == 2

        # dropping extra column not returned
        expected_cont = attr.evolve(cont, data=None)

        # but cont2 should be filtered out
        assert actual_result == [expected_cont]

    def test_content_find_murmur3_collision(self, swh_storage, mocker, sample_data):
        """The Murmur3 token is used as link from index tables to the main
        table; and non-matching contents with colliding murmur3-hash
        are filtered-out when reading the main table.
        This test checks the content methods do filter out these collisions.
        """
        called = 0

        cont, cont2 = [attr.evolve(c, ctime=now()) for c in sample_data.contents[:2]]

        # always return a token
        def mock_cgtfsa(algo, hashes):
            nonlocal called
            called += 1
            assert algo in ("sha1", "sha1_git")
            return [123456]

        mocker.patch.object(
            swh_storage._cql_runner,
            "content_get_tokens_from_single_algo",
            mock_cgtfsa,
        )

        # For all tokens, always return cont and cont2
        cols = list(set(cont.to_dict()) - {"data"})

        def mock_cgft(tokens):
            nonlocal called
            called += 1
            return [
                ContentRow(**{col: getattr(cont, col) for col in cols})
                for cont in [cont, cont2]
            ]

        mocker.patch.object(
            swh_storage._cql_runner, "content_get_from_tokens", mock_cgft
        )

        expected_content = attr.evolve(cont, data=None)

        actual_result = swh_storage.content_find({"sha1": cont.sha1})

        assert called == 2

        # but cont2 should be filtered out
        assert actual_result == [expected_content]

    def test_content_get_partition_murmur3_collision(
        self,
        swh_storage: CassandraStorage,
        mocker,
        sample_data,
    ):
        """The Murmur3 token is used as link from index tables to the main table; and
        non-matching contents with colliding murmur3-hash are filtered-out when reading
        the main table.

        This test checks the content_get_partition endpoints return all contents, even
        the collisions.

        """
        called = 0

        rows: Dict[int, Dict] = {}
        for tok, content in enumerate(sample_data.contents):
            cont = attr.evolve(content, data=None, ctime=now())
            row_d = {**cont.to_dict(), "tok": tok}
            rows[tok] = row_d

        # For all tokens, always return cont

        def mock_content_get_token_range(range_start, range_end, limit):
            nonlocal called
            called += 1

            for tok in list(rows.keys()) * 3:  # yield multiple times the same tok
                row_d = dict(rows[tok].items())
                row_d.pop("tok")
                yield (tok, ContentRow(**row_d))

        mocker.patch.object(
            swh_storage._cql_runner,
            "content_get_token_range",
            mock_content_get_token_range,
        )

        actual_results = list(
            stream_results(
                swh_storage.content_get_partition, partition_id=0, nb_partitions=1
            )
        )

        assert called > 0

        # everything is listed, even collisions
        assert len(actual_results) == 3 * len(sample_data.contents)
        # as we duplicated the returned results, dropping duplicate should yield
        # the original length
        assert len(set(actual_results)) == len(sample_data.contents)

    @pytest.mark.skip("content_update is not yet implemented for Cassandra")
    def test_content_update(self):
        pass

    def test_extid_murmur3_collision(self, swh_storage, mocker, sample_data):
        """The Murmur3 token is used as link from index table to the main
        table; and non-matching extid with colliding murmur3-hash
        are filtered-out when reading the main table.
        This test checks the extid methods do filter out these collision.
        """
        swh_storage.extid_add(sample_data.extids)

        # For any token, always return all extids, i.e. make as if all tokens
        # for all extid entries collide
        def mock_egft(token):
            return [
                ExtIDRow(
                    extid_type=extid.extid_type,
                    extid=extid.extid,
                    extid_version=extid.extid_version,
                    target_type=extid.target.object_type.value,
                    target=extid.target.object_id,
                )
                for extid in sample_data.extids
            ]

        mocker.patch.object(
            swh_storage._cql_runner,
            "extid_get_from_token",
            mock_egft,
        )

        for extid in sample_data.extids:
            extids = swh_storage.extid_get_from_target(
                target_type=extid.target.object_type, ids=[extid.target.object_id]
            )
            assert extids == [extid]

    def _directory_with_entries(self, sample_data, nb_entries):
        """Returns a dir with ``nb_entries``, all pointing to
        the same content"""
        return Directory(
            entries=tuple(
                DirectoryEntry(
                    name=f"file{i:10}".encode(),
                    type="file",
                    target=sample_data.content.sha1_git,
                    perms=from_disk.DentryPerms.directory,
                )
                for i in range(nb_entries)
            )
        )

    @pytest.mark.parametrize(
        "insert_algo,nb_entries",
        [
            ("one-by-one", 10),
            ("concurrent", 10),
            ("batch", 1),
            ("batch", 2),
            ("batch", BATCH_INSERT_MAX_SIZE - 1),
            ("batch", BATCH_INSERT_MAX_SIZE),
            ("batch", BATCH_INSERT_MAX_SIZE + 1),
            ("batch", BATCH_INSERT_MAX_SIZE * 2),
        ],
    )
    def test_directory_add_algos(
        self,
        swh_storage,
        sample_data,
        mocker,
        insert_algo,
        nb_entries,
    ):
        mocker.patch.object(swh_storage, "_directory_entries_insert_algo", insert_algo)

        class new_sample_data:
            content = sample_data.content
            directory = self._directory_with_entries(sample_data, nb_entries)

        self.test_directory_add(swh_storage, new_sample_data)

    @pytest.mark.parametrize("insert_algo", DIRECTORY_ENTRIES_INSERT_ALGOS)
    def test_directory_add_atomic(self, swh_storage, sample_data, mocker, insert_algo):
        """Checks that a crash occurring after some directory entries were written
        does not cause the directory to be (partially) visible.
        ie. checks directories are added somewhat atomically."""
        # Disable the journal writer, it would detect the CrashyEntry exception too
        # early for this test to be relevant
        swh_storage.journal_writer.journal = None
        mocker.patch.object(swh_storage, "_directory_entries_insert_algo", insert_algo)

        class CrashyEntry(DirectoryEntry):
            def __init__(self):
                super().__init__(**{**directory.entries[0].to_dict(), "name": b"crash"})

            def to_dict(self):
                return {**super().to_dict(), "perms": "abcde"}

        directory = self._directory_with_entries(sample_data, BATCH_INSERT_MAX_SIZE)
        entries = directory.entries
        directory = attr.evolve(directory, entries=entries + (CrashyEntry(),))

        with pytest.raises(TypeError):
            swh_storage.directory_add([directory])

        # Usually, this writes all entries but the crashy one in the database;
        # let's check this. (If this assertion fails, then the test is useless;
        # but it does not affect the actual functionality)
        # However, because they are inserted simultaneously, the backend may crash
        # before the last handful of entries; so we allow them to be missing
        # without failing the test (which would make it flaky).
        entry_rows = swh_storage._cql_runner.directory_entry_get([directory.id])
        assert (
            {entry.name for entry in entries[0:-100]}
            <= {row.name for row in entry_rows}
            <= {entry.name for entry in entries}
        )

        # BUT, because not all the entries were written, the directory should
        # be considered not written.
        assert swh_storage.directory_missing([directory.id]) == [directory.id]
        assert list(swh_storage.directory_ls(directory.id)) == []
        assert swh_storage.directory_get_entries(directory.id) is None

    def test_directory_add_raw_manifest__different_entries__allow_overwrite(
        self, swh_storage
    ):
        """This test demonstrates a shortcoming of the Cassandra storage backend's
        design:

        1. add a directory with an entry named "name1" and raw_manifest="abc"
        2. add a directory with an entry named "name2" and the same raw_manifest
        3. the directories' id is computed only from the raw_manifest, so both
           directories have the same id, which causes their entries to be
           "additive" in the database; so directory_ls returns both entries

        However, by default, the Cassandra storage has allow_overwrite=False,
        which "accidentally" avoids this issue most of the time, by skipping insertion
        if an object with the same id is already in the database.

        This can still be an issue when either allow_overwrite=True or when inserting
        both directories at about the same time (because of the lack of
        transactionality); but the likelihood of two clients inserting two different
        objects with the same manifest at the same time is very low, it could only
        happen if loaders running in parallel used different (or nondeterministic)
        parsers on corrupt objects.
        """
        assert (
            swh_storage._allow_overwrite is False
        ), "Unexpected default _allow_overwrite value"
        swh_storage._allow_overwrite = True

        # Run the other test, but skip its last assertion
        dir_id = self.test_directory_add_raw_manifest__different_entries(
            swh_storage, check_ls=False
        )
        assert [entry["name"] for entry in swh_storage.directory_ls(dir_id)] == [
            b"name1",
            b"name2",
        ]

    def test_snapshot_add_atomic(self, swh_storage, sample_data, mocker):
        """Checks that a crash occurring after some snapshot branches were written
        does not cause the snapshot to be (partially) visible.
        ie. checks snapshots are added somewhat atomically."""
        # Disable the journal writer, it would detect the CrashyBranch exception too
        # early for this test to be relevant
        swh_storage.journal_writer.journal = None

        class MyException(Exception):
            pass

        class CrashyBranch(SnapshotBranch):
            def __getattribute__(self, name):
                if name == "target" and should_raise:
                    raise MyException()
                else:
                    return super().__getattribute__(name)

        snapshot = sample_data.complete_snapshot
        branches = snapshot.branches

        should_raise = False  # just so that we can construct the object
        crashy_branch = CrashyBranch.from_dict(branches[b"directory"].to_dict())
        should_raise = True

        snapshot = attr.evolve(
            snapshot,
            branches={
                **branches,
                b"crashy": crashy_branch,
            },
        )

        with pytest.raises(MyException):
            swh_storage.snapshot_add([snapshot])

        # This should have written some of the branches to the database:
        branch_rows = swh_storage._cql_runner.snapshot_branch_get(snapshot.id, b"", 10)
        assert {row.name for row in branch_rows} == set(branches)

        # BUT, because not all the branches were written, the snapshot should
        # be considered not written.
        assert swh_storage.snapshot_missing([snapshot.id]) == [snapshot.id]
        assert swh_storage.snapshot_get(snapshot.id) is None
        assert swh_storage.snapshot_count_branches(snapshot.id) is None
        assert swh_storage.snapshot_get_branches(snapshot.id) is None

    @pytest.mark.skip(
        'The "person" table of the pgsql is a legacy thing, and not '
        "supported by the cassandra backend."
    )
    def test_person_fullname_unicity(self):
        pass

    @pytest.mark.skip(
        'The "person" table of the pgsql is a legacy thing, and not '
        "supported by the cassandra backend."
    )
    def test_person_get(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count(self):
        pass

    def test_object_delete(self, swh_storage, sample_data):
        # For our sanity checks
        affected_tables = set(TABLES) - {
            "metadata_authority",
            "metadata_fetcher",
            "extid",
            "extid_by_target",
            "object_references",
        }
        # XXX: Not ideal…
        cql_runner = getattr(swh_storage, "_cql_runner")
        execute_query = getattr(cql_runner, "_execute_with_retries")

        swh_storage.content_add(sample_data.contents)
        swh_storage.skipped_content_add(sample_data.skipped_contents)
        swh_storage.directory_add(sample_data.directories)
        swh_storage.revision_add(sample_data.revisions)
        swh_storage.release_add(sample_data.releases)
        swh_storage.snapshot_add(sample_data.snapshots)
        swh_storage.origin_add(sample_data.origins)
        swh_storage.origin_visit_add(sample_data.origin_visits)
        swh_storage.origin_visit_status_add(sample_data.origin_visit_statuses)
        swh_storage.metadata_authority_add(sample_data.authorities)
        swh_storage.metadata_fetcher_add(sample_data.fetchers)
        swh_storage.raw_extrinsic_metadata_add(sample_data.content_metadata)
        swh_storage.raw_extrinsic_metadata_add(sample_data.origin_metadata)
        swhids = (
            [content.swhid().to_extended() for content in sample_data.contents]
            + [
                skipped_content.swhid().to_extended()
                for skipped_content in sample_data.skipped_contents
            ]
            + [directory.swhid().to_extended() for directory in sample_data.directories]
            + [revision.swhid().to_extended() for revision in sample_data.revisions]
            + [release.swhid().to_extended() for release in sample_data.releases]
            + [snapshot.swhid().to_extended() for snapshot in sample_data.snapshots]
            + [origin.swhid() for origin in sample_data.origins]
            + [emd.swhid() for emd in sample_data.content_metadata]
            + [emd.swhid() for emd in sample_data.origin_metadata]
        )

        # Do we have something in every affected tables?
        for table in affected_tables:
            row = execute_query(
                f"SELECT COUNT(*) AS count FROM {cql_runner.keyspace}.{table}", []
            ).one()
            assert row["count"] >= 1, f"nothing in table {table}"

        result = swh_storage.object_delete(swhids)
        assert result == {
            "content:delete": 3,
            "content:delete:bytes": 0,
            "skipped_content:delete": 2,
            "directory:delete": 7,
            "revision:delete": 8,
            "release:delete": 3,
            "snapshot:delete": 3,
            "origin:delete": 7,
            "origin_visit:delete": 3,
            "origin_visit_status:delete": 3,
            "cnt_metadata:delete": 3,
            "ori_metadata:delete": 3,
        }

        # Have we cleaned every affected tables?
        for table in affected_tables:
            row = execute_query(
                f"SELECT COUNT(*) AS count FROM {cql_runner.keyspace}.{table}", []
            ).one()
            assert row["count"] == 0, f"something in table {table}"

    def test_extid_delete_for_target(self, swh_storage, sample_data):
        swh_storage.revision_add([sample_data.revision, sample_data.hg_revision])
        swh_storage.directory_add([sample_data.directory, sample_data.directory2])
        extid_for_same_target = ExtID(
            target=CoreSWHID(
                object_type=ObjectType.REVISION, object_id=sample_data.revision.id
            ),
            extid_type="drink_some",
            extid=bytes.fromhex("c0ffee"),
        )
        result = swh_storage.extid_add(sample_data.extids + (extid_for_same_target,))
        assert result == {"extid:add": 5}

        result = swh_storage.extid_delete_for_target(
            [sample_data.revision.swhid(), sample_data.directory2.swhid()]
        )
        assert result == {"extid:delete": 3}

        extids = swh_storage.extid_get_from_target(
            target_type=ObjectType.REVISION, ids=[sample_data.hg_revision.id]
        )
        assert extids == [sample_data.extid2]
        extids = swh_storage.extid_get_from_target(
            target_type=ObjectType.DIRECTORY, ids=[sample_data.directory.id]
        )
        assert extids == [sample_data.extid3]

        result = swh_storage.extid_delete_for_target(
            [sample_data.hg_revision.swhid(), sample_data.directory.swhid()]
        )
        assert result == {"extid:delete": 2}


@pytest.mark.cassandra
class TestCassandraStorageGeneratedData(_TestStorageGeneratedData):
    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_no_visits(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_with_visits_and_snapshot(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_with_visits_no_snapshot(self):
        pass


@pytest.mark.cassandra
@pytest.mark.parametrize(
    "allow_overwrite,object_type",
    itertools.product(
        [False, True],
        # Note the absence of "content", it's tested above.
        ["directory", "revision", "release", "snapshot", "origin", "extid"],
    ),
)
def test_allow_overwrite(
    allow_overwrite: bool, object_type: str, swh_storage_backend_config
):
    if object_type in ("origin", "extid"):
        pytest.skip(
            f"test_disallow_overwrite not implemented for {object_type} objects, "
            f"because all their columns are in the primary key."
        )
    swh_storage = get_storage(
        allow_overwrite=allow_overwrite, **swh_storage_backend_config
    )

    # directory_ls joins with content and directory table, and needs those to return
    # non-None entries:
    if object_type == "directory":
        swh_storage.directory_add([StorageData.directory5])
        swh_storage.content_add([StorageData.content, StorageData.content2])

    obj1: Any
    obj2: Any

    # Get two test objects
    if object_type == "directory":
        (obj1, obj2, *_) = StorageData.directories
    elif object_type == "snapshot":
        # StorageData.snapshots[1] is the empty snapshot, which is the corner case
        # that makes this test succeed for the wrong reasons
        obj1 = StorageData.snapshot
        obj2 = StorageData.complete_snapshot
    else:
        (obj1, obj2, *_) = getattr(StorageData, (object_type + "s"))

    # Let's make both objects have the same hash, but different content
    obj1 = attr.evolve(obj1, id=obj2.id)

    # Get the methods used to add and get these objects
    add = getattr(swh_storage, object_type + "_add")
    if object_type == "directory":

        def get(ids):
            return [
                Directory(
                    id=ids[0],
                    entries=tuple(
                        map(
                            lambda entry: DirectoryEntry(
                                name=entry["name"],
                                type=entry["type"],
                                target=entry["sha1_git"],
                                perms=entry["perms"],
                            ),
                            swh_storage.directory_ls(ids[0]),
                        )
                    ),
                )
            ]

    elif object_type == "snapshot":

        def get(ids):
            return [
                Snapshot.from_dict(
                    remove_keys(swh_storage.snapshot_get(ids[0]), ("next_branch",))
                )
            ]

    else:
        get = getattr(swh_storage, object_type + "_get")

    # Add the first object
    add([obj1])

    # It should be returned as-is
    assert get([obj1.id]) == [obj1]

    # Add the second object
    add([obj2])

    if allow_overwrite:
        # obj1 was overwritten by obj2
        expected = obj2
    else:
        # obj2 was not written, because obj1 already exists and has the same hash
        expected = obj1

    if allow_overwrite and object_type in ("directory", "snapshot"):
        # TODO
        pytest.xfail(
            "directory entries and snapshot branches are concatenated "
            "instead of being replaced"
        )
    assert get([obj1.id]) == [expected]



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import deepcopy
import datetime
import json
from typing import Dict, Tuple

import attr

from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    ObjectType,
    OriginVisit,
    OriginVisitStatus,
    RawExtrinsicMetadata,
    Release,
    Revision,
    RevisionType,
    Sha1Git,
)
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID
from swh.storage.interface import ObjectReference

from ..utils import remove_keys
from .model import (
    ObjectReferenceRow,
    OriginVisitRow,
    OriginVisitStatusRow,
    RawExtrinsicMetadataRow,
    ReleaseRow,
    RevisionRow,
)


def revision_to_db(revision: Revision) -> RevisionRow:
    # we use a deepcopy of the dict because we do not want to recurse the
    # Model->dict conversion (to keep Timestamp & al. entities), BUT we do not
    # want to modify original metadata (embedded in the Model entity), so we
    # non-recursively convert it as a dict but make a deep copy.
    db_revision = deepcopy(attr.asdict(revision, recurse=False))
    metadata = revision.metadata
    extra_headers = revision.extra_headers
    if not extra_headers and metadata and "extra_headers" in metadata:
        extra_headers = db_revision["metadata"].pop("extra_headers")
    db_revision["metadata"] = json.dumps(
        dict(db_revision["metadata"]) if db_revision["metadata"] is not None else None
    )
    db_revision["extra_headers"] = extra_headers
    db_revision["type"] = db_revision["type"].value
    return RevisionRow(**remove_keys(db_revision, ("parents",)))


def revision_from_db(
    db_revision: RevisionRow, parents: Tuple[Sha1Git, ...]
) -> Revision:
    revision = db_revision.to_dict()
    metadata = json.loads(revision.pop("metadata", None))
    extra_headers = revision.pop("extra_headers", ())
    if not extra_headers and metadata and "extra_headers" in metadata:
        extra_headers = metadata.pop("extra_headers")
    if extra_headers is None:
        extra_headers = ()
    return Revision(
        parents=parents,
        type=RevisionType(revision.pop("type")),
        metadata=metadata,
        extra_headers=extra_headers,
        **revision,
    )


def release_to_db(release: Release) -> ReleaseRow:
    db_release = attr.asdict(release, recurse=False)
    db_release["target_type"] = db_release["target_type"].value
    return ReleaseRow(**remove_keys(db_release, ("metadata",)))


def release_from_db(db_release: ReleaseRow) -> Release:
    release = db_release.to_dict()
    return Release(
        target_type=ObjectType(release.pop("target_type")),
        **release,
    )


def row_to_content_hashes(row: ReleaseRow) -> Dict[str, bytes]:
    """Convert cassandra row to a content hashes"""
    hashes = {}
    for algo in DEFAULT_ALGORITHMS:
        hashes[algo] = getattr(row, algo)
    return hashes


def row_to_visit(row: OriginVisitRow) -> OriginVisit:
    """Format a row representing an origin_visit to an actual OriginVisit."""
    return OriginVisit(
        origin=row.origin,
        visit=row.visit,
        date=row.date.replace(tzinfo=datetime.timezone.utc),
        type=row.type,
    )


def row_to_visit_status(row: OriginVisitStatusRow) -> OriginVisitStatus:
    """Format a row representing a visit_status to an actual OriginVisitStatus."""
    return OriginVisitStatus.from_dict(
        {
            **row.to_dict(),
            "date": row.date.replace(tzinfo=datetime.timezone.utc),
            "metadata": (json.loads(row.metadata) if row.metadata else None),
        }
    )


def visit_status_to_row(status: OriginVisitStatus) -> OriginVisitStatusRow:
    d = status.to_dict()
    return OriginVisitStatusRow.from_dict({**d, "metadata": json.dumps(d["metadata"])})


def row_to_raw_extrinsic_metadata(row: RawExtrinsicMetadataRow) -> RawExtrinsicMetadata:
    discovery_date = row.discovery_date.replace(tzinfo=datetime.timezone.utc)

    return RawExtrinsicMetadata(
        target=ExtendedSWHID.from_string(row.target),
        authority=MetadataAuthority(
            type=MetadataAuthorityType(row.authority_type),
            url=row.authority_url,
        ),
        fetcher=MetadataFetcher(
            name=row.fetcher_name,
            version=row.fetcher_version,
        ),
        discovery_date=discovery_date,
        format=row.format,
        metadata=row.metadata,
        origin=row.origin or None,  # to account for "" conversion to None
        visit=row.visit if row.visit and row.visit != 0 else None,
        snapshot=CoreSWHID.from_string(row.snapshot) if row.snapshot else None,
        release=CoreSWHID.from_string(row.release) if row.release else None,
        revision=CoreSWHID.from_string(row.revision) if row.revision else None,
        path=row.path or None,  # to account for b"" conversion to None
        directory=CoreSWHID.from_string(row.directory) if row.directory else None,
    )


def object_reference_to_row(object_reference: ObjectReference) -> ObjectReferenceRow:
    return ObjectReferenceRow(
        source_type=object_reference.source.object_type.value,
        source=object_reference.source.object_id,
        target_type=object_reference.target.object_type.value,
        target=object_reference.target.object_id,
    )


def row_to_object_reference(row: ObjectReferenceRow) -> ObjectReference:
    return ObjectReference(
        source=ExtendedSWHID(
            object_type=ExtendedObjectType(row.source_type), object_id=row.source
        ),
        target=ExtendedSWHID(
            object_type=ExtendedObjectType(row.target_type), object_id=row.target
        ),
    )



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Any, Callable, Dict, List

from swh.model.model import Origin

logger = logging.getLogger(__name__)


def _fix_content(content: Dict[str, Any]) -> Dict[str, Any]:
    """Filters-out invalid 'perms' key that leaked from swh.model.from_disk
    to the journal.

    >>> _fix_content({'perms': 0o100644, 'sha1_git': b'foo'})
    {'sha1_git': b'foo'}

    >>> _fix_content({'sha1_git': b'bar'})
    {'sha1_git': b'bar'}

    """
    content = content.copy()
    content.pop("perms", None)
    return content


def _fix_raw_extrinsic_metadata(obj_dict: Dict) -> Dict:
    """Fix legacy RawExtrinsicMetadata with type which is no longer part of the model.

    >>> _fix_raw_extrinsic_metadata({
    ...     'type': 'directory',
    ...     'target': 'swh:1:dir:460a586d1c95d120811eaadb398d534e019b5243',
    ... })
    {'target': 'swh:1:dir:460a586d1c95d120811eaadb398d534e019b5243'}
    >>> _fix_raw_extrinsic_metadata({
    ...     'type': 'origin',
    ...     'target': 'https://inria.halpreprod.archives-ouvertes.fr/hal-01667309',
    ... })
    {'target': 'swh:1:ori:155291d5b9ada4570672510509f93fcfd9809882'}

    """
    o = obj_dict.copy()
    if o.pop("type", None) == "origin":
        o["target"] = str(Origin(o["target"]).swhid())
    return o


object_fixers: Dict[str, Callable[[Dict], Dict]] = {
    "content": _fix_content,
    "raw_extrinsic_metadata": _fix_raw_extrinsic_metadata,
}


def fix_objects(object_type: str, objects: List[Dict]) -> List[Dict]:
    """
    Fix legacy objects from the journal to bring them up to date with the
    latest storage schema.
    """
    if object_type in object_fixers:
        fixer = object_fixers[object_type]
        objects = [fixer(v) for v in objects]
    return objects



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import datetime
import logging
import os
import pathlib
import re
import tempfile
from unittest.mock import patch

from click.testing import CliRunner
from confluent_kafka import Producer
import pytest
import yaml

from swh.journal.serializers import key_to_kafka, value_to_kafka
from swh.model.model import Origin, Snapshot, SnapshotBranch, SnapshotTargetType
from swh.storage import get_storage
from swh.storage.cli import storage as cli
from swh.storage.replay import OBJECT_CONVERTERS

logger = logging.getLogger(__name__)


CLI_CONFIG = {
    "storage": {
        "cls": "memory",
    },
}


@pytest.fixture
def swh_storage():
    """An swh-storage object that gets injected into the CLI functions."""
    storage = get_storage(**CLI_CONFIG["storage"])
    with patch("swh.storage.get_storage") as get_storage_mock:
        get_storage_mock.return_value = storage
        yield storage


def invoke(*args, env=None, input=None, journal_config=None, local_config=None):
    config = local_config or copy.deepcopy(CLI_CONFIG)
    if journal_config:
        config["journal_client"] = journal_config.copy()
        config["journal_client"]["cls"] = "kafka"

    runner = CliRunner()
    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        yaml.dump(config, config_fd)
        config_fd.seek(0)
        args = ["-C" + config_fd.name] + list(args)
        ret = runner.invoke(
            cli,
            args,
            obj={"log_level": logging.DEBUG},
            env=env,
            input=input,
        )
        return ret


@pytest.mark.cassandra
def test_create_keyspace(
    swh_storage_cassandra_cluster,
    cassandra_auth_provider_config,
):
    (hosts, port) = swh_storage_cassandra_cluster
    keyspace = "test" + os.urandom(10).hex()

    storage_config = dict(
        cls="cassandra",
        hosts=hosts,
        port=port,
        keyspace=keyspace,
        journal_writer={"cls": "memory"},
        objstorage={"cls": "memory"},
        auth_provider=cassandra_auth_provider_config,
    )

    result = invoke("create-keyspace", local_config={"storage": storage_config})
    assert result.exit_code == 0, result.output
    assert result.output == "Done.\n"

    # Check we can write and read to it
    storage = get_storage(**storage_config)
    origin = Origin(url="http://example.org")
    storage.origin_add([origin])
    assert storage.origin_get([origin.url]) == [origin]


def test_replay(
    swh_storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
):
    kafka_prefix += ".swh.journal.objects"

    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test-producer",
            "acks": "all",
        }
    )

    snapshot = Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target_type=SnapshotTargetType.REVISION,
                target=b"\x01" * 20,
            )
        },
    )
    snapshot_dict = snapshot.to_dict()

    producer.produce(
        topic=kafka_prefix + ".snapshot",
        key=key_to_kafka(snapshot.id),
        value=value_to_kafka(snapshot_dict),
    )
    producer.flush()

    logger.debug("Flushed producer")

    result = invoke(
        "replay",
        "--stop-after-objects",
        "1",
        journal_config={
            "brokers": [kafka_server],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
        },
    )

    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    assert swh_storage.snapshot_get(snapshot.id) == {
        **snapshot_dict,
        "next_branch": None,
    }


def test_replay_with_exceptions(
    swh_storage,
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
    tmp_path: pathlib.Path,
):
    kafka_prefix += ".swh.journal.objects"

    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test-producer",
            "acks": "all",
        }
    )

    snapshot = Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target_type=SnapshotTargetType.REVISION,
                target=b"\x01" * 20,
            )
        },
    )
    snapshot_dict = snapshot.to_dict()
    producer.produce(
        topic=kafka_prefix + ".snapshot",
        key=key_to_kafka(snapshot.id),
        value=value_to_kafka(snapshot_dict),
    )
    # add 2 invalid snapshots
    inv_id = b"\x00" * 20
    inv_snapshot = {**snapshot_dict, "id": inv_id}
    producer.produce(
        topic=kafka_prefix + ".snapshot",
        key=key_to_kafka(inv_id),
        value=value_to_kafka(inv_snapshot),
    )
    inv_id2 = b"\x02" * 20
    inv_snapshot2 = {**snapshot_dict, "id": inv_id2}
    producer.produce(
        topic=kafka_prefix + ".snapshot",
        key=key_to_kafka(inv_id2),
        value=value_to_kafka(inv_snapshot2),
    )

    producer.flush()

    logger.debug("Flushed producer")

    # make an exception for the the first invalid snp only
    excfile = tmp_path / "swhids.txt"
    excfile.write_text(
        f"swh:1:snp:0000000000000000000000000000000000000000,{snapshot.id.hex()}\n\n"
    )

    result = invoke(
        "replay",
        "--stop-after-objects",
        "3",
        "--known-mismatched-hashes",
        str(excfile.absolute()),
        journal_config={
            "brokers": [kafka_server],
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
            "privileged": True,  # required to activate validation
        },
    )

    expected = r"Done.\n"
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    assert swh_storage.snapshot_get(snapshot.id) == dict(
        **snapshot_dict,
        next_branch=None,
    )
    assert swh_storage.snapshot_get(inv_id) == dict(
        **inv_snapshot,
        next_branch=None,
    )
    assert swh_storage.snapshot_get(inv_id2) is None


def test_replay_type_list():
    result = invoke(
        "replay",
        "--help",
    )
    assert result.exit_code == 0, result.output
    types_in_help = re.findall("--type [[]([a-z_|]+)[]]", result.output)
    assert len(types_in_help) == 1
    types = types_in_help[0].split("|")

    assert sorted(types) == sorted(list(OBJECT_CONVERTERS.keys())), (
        "Make sure the list of accepted types in cli.py "
        "matches implementation in replay.py"
    )


@pytest.mark.parametrize(
    ("start", "end", "expected_weeks", "unexpected_weeks"),
    (
        pytest.param(
            "2020-02-03",
            "2020-02-09",
            [("2020w06", "2020-02-03", "2020-02-10")],
            ["2020w05", "2020w07"],
            id="single-week",
        ),
        pytest.param(
            "2023-01-01",
            "2023-01-01",
            [("2022w52", "2022-12-26", "2023-01-02")],
            ["2022w51", "2023w01"],
            id="week-in-another-year",
        ),
        pytest.param(
            "2023-01-01",
            "2023-01-17",
            [
                ("2022w52", "2022-12-26", "2023-01-02"),
                ("2023w01", "2023-01-02", "2023-01-09"),
                ("2023w02", "2023-01-09", "2023-01-16"),
                ("2023w03", "2023-01-16", "2023-01-23"),
            ],
            ["2022w51", "2023w04"],
            id="multiple-weeks",
        ),
    ),
)
def test_create_object_reference_partitions_postgresql(
    swh_storage_postgresql_backend_config, start, end, expected_weeks, unexpected_weeks
):
    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_postgresql_backend_config,
            ],
        }
    }

    result = invoke(
        "create-object-reference-partitions",
        start,
        end,
        local_config=storage_config,
    )

    assert result.exit_code == 0, result.output

    swh_storage = get_storage(**swh_storage_postgresql_backend_config)

    with swh_storage.db() as db:
        partitions = db.object_references_list_partitions()

    for partition, (week, expected_start, expected_end) in zip(
        partitions, expected_weeks
    ):
        assert partition.table_name == f"object_references_{week}"
        assert partition.start == datetime.datetime.fromisoformat(expected_start)
        assert partition.end == datetime.datetime.fromisoformat(expected_end)

    table_names = {partition.table_name for partition in partitions}
    for week in unexpected_weeks:
        assert f"object_references_{week}" not in table_names


@pytest.fixture
def swh_storage_with_partitions(swh_storage_postgresql_backend_config):
    # Setup partitions from 2022-12-26 00:00 to 2023-01-15 23:59
    swh_storage = get_storage(**swh_storage_postgresql_backend_config)
    with swh_storage.db() as db:
        with db.transaction() as cur:
            db.object_references_create_partition(2022, 52, cur)
            for week in range(1, 3):
                db.object_references_create_partition(2023, week, cur)
    return swh_storage


@pytest.mark.parametrize(
    ("before", "listed_in_output", "remaining"),
    (
        pytest.param(
            "2022-12-25",
            [],
            ["2022w52", "2023w01", "2023w02"],
            id="older-than-first-week",
        ),
        pytest.param(
            "2023-01-01",
            [],
            ["2022w52", "2023w01", "2023w02"],
            id="sunday-of-first-week",
        ),
        pytest.param(
            "2023-01-02",
            ["2022 week 52"],
            ["2023w01", "2023w02"],
            id="middle-of-second-week",
        ),
        pytest.param(
            "2023-01-12",
            ["2022 week 52", "2023 week 01"],
            ["2023w02"],
            id="middle-of-third-week",
        ),
    ),
)
def test_remove_old_object_reference_partitions_postgresql(
    swh_storage_postgresql_backend_config,
    swh_storage_with_partitions,
    before,
    listed_in_output,
    remaining,
):
    from datetime import datetime

    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_postgresql_backend_config,
            ],
        }
    }

    result = invoke(
        "remove-old-object-reference-partitions",
        before,
        input="y\n",
        local_config=storage_config,
    )

    assert result.exit_code == 0, result.output

    assert all(s in result.output for s in listed_in_output)

    with swh_storage_with_partitions.db() as db:
        partitions = db.object_references_list_partitions()

    table_names = {partition.table_name for partition in partitions}

    # We get a partition created for the current week when the schema is initialized
    table_names.discard(f"object_references_{datetime.now().strftime('%Yw%W')}")
    assert {f"object_references_{week}" for week in remaining} == table_names


def test_remove_old_object_reference_partitions_postgresql_refuses_to_remove_all(
    swh_storage_postgresql_backend_config, swh_storage_with_partitions
):
    from datetime import datetime, timedelta

    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_postgresql_backend_config,
            ],
        }
    }

    result = invoke(
        "remove-old-object-reference-partitions",
        # A partition for this week is created with the schema,
        # so we just ask to delete
        (datetime.now() + timedelta(weeks=1)).strftime("%Y-%m-%d"),
        local_config=storage_config,
    )

    assert result.exit_code != 0
    assert "Trying to remove all existing partitions" in result.output

    with swh_storage_with_partitions.db() as db:
        partitions = db.object_references_list_partitions()

    # The three partition created in the fixture + the one created when the
    # schema was initialized.
    assert len(partitions) == 4


def test_remove_old_object_reference_partitions_postgresql_using_force_argument(
    swh_storage_postgresql_backend_config, swh_storage_with_partitions
):
    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_postgresql_backend_config,
            ],
        }
    }

    result = invoke(
        "remove-old-object-reference-partitions",
        "--force",
        "2023-01-31",
        local_config=storage_config,
    )

    assert result.exit_code == 0, result.output

    with swh_storage_with_partitions.db() as db:
        partitions = db.object_references_list_partitions()

    table_names = {partition.table_name for partition in partitions}

    assert not any(
        f"object_references_{week}" in table_names
        for week in ("2023w01", "2023w02", "2023w03")
    )



# Copyright (C) 2020-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, Iterable, List, Optional, Tuple, Union, cast
import warnings

from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.model import Content, MissingData, Sha1
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import CompositeObjId
from swh.storage.interface import HashDict, StorageInterface

from .exc import StorageArgumentException


class ObjStorage:
    """Objstorage collaborator in charge of adding objects to
    the objstorage.

    """

    def __init__(self, storage: StorageInterface, objstorage_config: Optional[Dict]):
        self.storage = storage
        self.warn_usage = False
        if objstorage_config is None:
            objstorage_config = {"cls": "noop"}
            self.warn_usage = True
        self.objstorage = get_objstorage(**objstorage_config)

    def __getattr__(self, key):
        if key in ("objstorage", "warn_usage", "storage"):
            raise AttributeError(key)
        if self.warn_usage:
            warnings.warn(
                "Actually using a NoopObjstorage; this is most probably a configuration error.",
            )
        return getattr(self.objstorage, key)

    def content_get(self, obj_id: Union[Sha1, HashDict]) -> Optional[bytes]:
        """Retrieve data associated to the content from the objstorage

        Args:
            content: content identitier

        Returns:
            associated content's data if any, None otherwise.

        """
        if self.warn_usage:
            warnings.warn(
                "Actually using a NoopObjstorage; this is most probably a configuration error.",
            )
        hashes: HashDict
        if isinstance(obj_id, bytes):
            warnings.warn(
                'Identifying contents by sha1 instead of hash dicts `{"sha1": b"..."}` '
                "is deprecated.",
                DeprecationWarning,
                stacklevel=3,  # Report to the caller of swh/storage/*/storage.py
            )
            hashes = {"sha1": obj_id}
        else:
            hashes = obj_id
        if set(hashes) < DEFAULT_ALGORITHMS:
            # If some hashes are missing, query the database to fill blanks
            candidates = self.storage.content_find(hashes)
            if candidates:
                # There may be more than one in case of collision; but we cannot
                # do anything about it here
                hashes = cast(HashDict, candidates[0].hashes())
            else:
                # we will pass the partial hash dict to the objstorage, which
                # will do the best it can with it. Usually, this will return None,
                # as objects missing from the storage DB are unlikely to be present in the
                # objstorage
                pass
        try:
            data = self.objstorage.get(hashes)
        except ObjNotFoundError:
            data = None

        return data

    def content_add(self, contents: Iterable[Content]) -> Dict:
        """Add contents to the objstorage.

        Args:
            contents: List of contents to add1

        Returns:
            The summary dict of content and content bytes added to the
            objstorage.

        """
        if self.warn_usage:
            warnings.warn(
                "Actually using a NoopObjstorage; this is most probably a configuration error.",
            )
        content_pairs: List[Tuple[CompositeObjId, bytes]] = []
        for content in contents:
            try:
                content = content.with_data()
            except MissingData:
                raise StorageArgumentException("Missing data") from None
            assert content.data is not None
            content_pairs.append((cast(CompositeObjId, content.hashes()), content.data))
        summary = self.objstorage.add_batch(content_pairs)
        return {
            "content:add": summary["object:add"],
            "content:add:bytes": summary["object:add:bytes"],
        }



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
import functools
import inspect
import itertools
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)
import warnings

import attr
import psycopg2.pool

from swh.core.utils import grouper
from swh.model.hashutil import MultiHash
from swh.model.model import Origin, Person, Release, Revision, Sha1Git
from swh.model.swhids import ExtendedObjectType, ExtendedSWHID
from swh.storage import get_storage
from swh.storage.exc import MaskedObjectException
from swh.storage.interface import HashDict, PagedResult, Sha1, StorageInterface, TResult
from swh.storage.metrics import DifferentialTimer
from swh.storage.proxies.masking.db import MaskedStatus

from .db import MaskingQuery

BATCH_SIZE = 1024
MASKING_OVERHEAD_METRIC = "swh_storage_masking_overhead_seconds"


def get_datastore(cls, db=None, masking_db=None, **kwargs):
    assert cls in ("postgresql", "masking")
    from .db import MaskingAdmin

    if db is None:
        db = masking_db
    return MaskingAdmin.connect(db)


def masking_overhead_timer(method_name: str) -> DifferentialTimer:
    """Return a properly setup DifferentialTimer for ``method_name`` of the storage"""
    return DifferentialTimer(MASKING_OVERHEAD_METRIC, tags={"endpoint": method_name})


class MaskingProxyStorage:
    """Masking storage proxy

    This proxy can return modified objects or stop them from being retrieved at
    all.

    It uses a specific PostgreSQL database (which for now is colocated with the
    swh.storage PostgreSQL database), the access to which is implemented in the
    :mod:`.db` submodule.

    Sample configuration

    .. code-block: yaml

        storage:
          cls: masking
          db: 'dbname=swh-storage'
          max_pool_conns: 10
          storage:
          - cls: remote
            url: http://storage.internal.staging.swh.network:5002/

    """

    def __init__(
        self,
        storage: Union[Dict, StorageInterface],
        db: Optional[str] = None,
        masking_db: Optional[str] = None,
        min_pool_conns: int = 1,
        max_pool_conns: int = 5,
    ):
        if db is None:
            assert masking_db is not None
            warnings.warn(
                "'masking_db' field in the masking storage configuration "
                "was renamed 'db' field",
                DeprecationWarning,
            )
            db = masking_db
        self.storage: StorageInterface = (
            get_storage(**storage) if isinstance(storage, dict) else storage
        )

        self._masking_pool = psycopg2.pool.ThreadedConnectionPool(
            min_pool_conns, max_pool_conns, db
        )

        # Generate the method dictionaries once per instantiation, instead of
        # doing it on every (first) __getattr__ call.
        self._gen_method_dicts()

    @contextmanager
    def _masking_query(self) -> Iterator[MaskingQuery]:
        ret = None
        try:
            ret = MaskingQuery.from_pool(self._masking_pool)
            yield ret
        finally:
            if ret:
                ret.put_conn()

    @staticmethod
    def _get_swhids_in_result(method_name: str, result: Any) -> List[ExtendedSWHID]:
        if result is None:
            raise TypeError(f"Filtering of Nones missing in {method_name}")

        result_type = getattr(result, "object_type", None)
        if result_type == "raw_extrinsic_metadata":
            # Raw Extrinsic Metadata have a swhid, but we also mask them if the target is masked
            return [result.swhid(), result.target]

        if hasattr(result, "swhid"):
            swhid = result.swhid()
            if hasattr(swhid, "to_extended"):
                swhid = swhid.to_extended()
            assert isinstance(
                swhid, ExtendedSWHID
            ), f"{method_name} returned object with unexpected swhid() method return type"
            return [swhid]

        if result_type:
            if result_type == "extid":
                return [result.target.to_extended()]
            if result_type.value.startswith("origin_visit"):
                return [Origin(url=result.origin).swhid()]

        if (
            object_type := method_name.removesuffix("_get_random").removesuffix(
                "_get_id_partition"
            )
        ) != method_name:
            # Returns bare sha1_gits
            assert isinstance(result, bytes), f"{method_name} returned unexpected type"
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType[object_type.upper()],
                    object_id=result,
                )
            ]
        elif method_name == "revision_log":
            # returns dicts of revisions
            assert (
                isinstance(result, dict) and "id" in result
            ), f"{method_name} returned unexpected type"
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.REVISION, object_id=result["id"]
                )
            ]
        elif method_name == "revision_shortlog":
            # Returns tuples (revision.id, revision.parents)
            assert isinstance(
                result[0], bytes
            ), f"{method_name} returned unexpected type"
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.REVISION, object_id=result[0]
                )
            ]

        elif method_name == "origin_get_by_sha1":
            # Returns origin dicts, because why not
            assert (
                isinstance(result, dict) and "url" in result
            ), f"{method_name} returned unexpected type"
            return [Origin(url=result["url"]).swhid()]
        elif method_name == "origin_visit_get_with_statuses":
            # Returns an OriginVisitWithStatuses
            return [Origin(url=result.visit.origin).swhid()]
        elif method_name == "snapshot_get":
            # Returns a snapshot dict
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.SNAPSHOT, object_id=result["id"]
                )
            ]

        raise ValueError(f"Cannot get swhid for result of method {method_name}")

    def _masked_result(
        self, method_name: str, result: Any
    ) -> Optional[Dict[ExtendedSWHID, List[MaskedStatus]]]:
        """Find the SWHIDs of the ``result`` object, and check if any of them is
        masked, returning the associated masking information."""
        with self._masking_query() as q:
            return q.swhids_are_masked(self._get_swhids_in_result(method_name, result))

    def _raise_if_masked_result(self, method_name: str, result: Any) -> None:
        """Raise a :exc:`MaskedObjectException` if ``result`` is masked."""
        masked = self._masked_result(method_name, result)
        if masked:
            raise MaskedObjectException(masked)

    def _raise_if_masked_swhids(self, swhids: List[ExtendedSWHID]) -> None:
        """Raise a :exc:`MaskedObjectException` if any SWHID is masked."""
        with self._masking_query() as q:
            masked = q.swhids_are_masked(swhids)

        if masked:
            raise MaskedObjectException(masked)

    def __getattr__(self, key):
        method = None
        if key in self._methods_by_name:
            method = self._methods_by_name[key](key)
        else:
            suffix = key.rsplit("_", 1)[-1]

            if suffix in self._methods_by_suffix:
                method = self._methods_by_suffix[suffix](key)

        if method:
            # Avoid going through __getattr__ again next time
            setattr(self, key, method)
            return method

        # Raise a NotImplementedError to make sure we don't forget to add
        # masking to any new storage functions
        raise NotImplementedError(key)

    def _gen_method_dicts(self):
        """Generate the :attr:`_methods_by_name` and :attr:`_methods_by_suffix`
        used by :meth:`__getattr__`"""
        _passthrough = functools.partial(getattr, self.storage)

        self._methods_by_name = {
            # Returns a single object
            "snapshot_get": self._getter_optional,
            "origin_visit_find_by_date": self._getter_optional,
            "origin_visit_get_by": self._getter_optional,
            "origin_visit_status_get_latest": self._getter_optional,
            "origin_visit_get_latest": self._getter_optional,
            # Returns a PagedResult
            "origin_list": self._getter_pagedresult,
            "origin_visit_get": self._getter_pagedresult,
            "origin_search": self._getter_pagedresult,
            "raw_extrinsic_metadata_get": self._getter_pagedresult,
            "origin_visit_get_with_statuses": self._getter_pagedresult,
            "origin_visit_status_get": self._getter_pagedresult,
            # Returns a list of (optional) objects
            "origin_get_by_sha1": self._getter_list,
            "content_find": self._getter_list,
            "skipped_content_find": self._getter_list,
            "revision_shortlog": self._getter_list,
            "extid_get_from_target": self._getter_list,
            "raw_extrinsic_metadata_get_by_ids": self._getter_list,
            # Filter arguments
            "directory_entry_get_by_path": self._getter_filtering_arguments,
            "directory_get_entries": self._getter_filtering_arguments,
            "directory_get_raw_manifest": self._getter_filtering_arguments,
            "directory_ls": self._getter_filtering_arguments,
            "raw_extrinsic_metadata_get_authorities": self._getter_filtering_arguments,
            "snapshot_branch_get_by_name": self._getter_filtering_arguments,
            "snapshot_count_branches": self._getter_filtering_arguments,
            "snapshot_get_branches": self._getter_filtering_arguments,
            "origin_snapshot_get_all": self._getter_filtering_arguments,
            # Content functions that don't match common getter or adder suffixes
            "content_add_metadata": _passthrough,
            "content_missing_per_sha1": _passthrough,
            "content_missing_per_sha1_git": _passthrough,
            "content_update": _passthrough,
            # These objects aren't maskable
            "extid_get_from_extid": _passthrough,
            "object_find_by_sha1_git": _passthrough,
            "object_find_recent_references": _passthrough,
            "metadata_authority_get": _passthrough,
            "metadata_fetcher_get": _passthrough,
            # Utility methods
            "check_config": _passthrough,
            "clear_buffers": _passthrough,
            "flush": _passthrough,
            "origin_count": _passthrough,
            "refresh_stat_counters": _passthrough,
            "stat_counters": _passthrough,
            # For tests
            "journal_writer": _passthrough,
        }

        self._methods_by_suffix = {
            # These methods will never need do any masking
            "add": _passthrough,
            "missing": _passthrough,
            # Partitions return PagedResults
            "partition": self._getter_pagedresult,
            # Getters return lists of optional objects
            "get": self._getter_list,
            "random": self._getter_random,
        }

    def content_get_data(self, content: Union[HashDict, Sha1]) -> Optional[bytes]:
        with masking_overhead_timer("content_get_data") as t:
            with t.inner():
                ret = self.storage.content_get_data(content)

            if ret is None:
                return None

            if isinstance(content, dict) and "sha1_git" in content:
                self._raise_if_masked_swhids(
                    [
                        ExtendedSWHID(
                            object_type=ExtendedObjectType.CONTENT,
                            object_id=content["sha1_git"],
                        )
                    ]
                )

            else:
                # We did not get the SWHID of the object as argument, so we need to
                # hash the resulting content to check if its SWHID was masked.
                self._raise_if_masked_swhids(
                    [
                        ExtendedSWHID(
                            object_type=ExtendedObjectType.CONTENT,
                            object_id=MultiHash.from_data(ret, ["sha1_git"]).digest()[
                                "sha1_git"
                            ],
                        )
                    ]
                )

            return ret

    def _get_swhids_in_args(
        self, method_name: str, parsed_args: Dict[str, Any]
    ) -> List[ExtendedSWHID]:
        """Extract SWHIDs from the parsed arguments of ``method_name``.

        Arguments:
          method_name: name of the called method
          parsed_args: arguments of the method parsed with :func:`inspect.getcallargs`
        """

        if method_name in ("directory_entry_get_by_path", "directory_ls"):
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.DIRECTORY,
                    object_id=parsed_args["directory"],
                )
            ]
        elif method_name == "directory_get_entries":
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.DIRECTORY,
                    object_id=parsed_args["directory_id"],
                )
            ]
        elif method_name.startswith("snapshot_"):
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.SNAPSHOT,
                    object_id=parsed_args["snapshot_id"],
                )
            ]
        elif method_name == "raw_extrinsic_metadata_get_authorities":
            return [parsed_args["target"]]
        elif method_name == "directory_get_raw_manifest":
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.DIRECTORY,
                    object_id=object_id,
                )
                for object_id in parsed_args["directory_ids"]
            ]
        elif method_name == "origin_snapshot_get_all":
            return [Origin(url=parsed_args["origin_url"]).swhid()]
        else:
            raise ValueError(f"Cannot get swhid for arguments of method {method_name}")

    def _getter_filtering_arguments(self, method_name: str):
        """Handles methods that should filter on their argument, instead of the
        returned value. If the underlying storage returns :const:`None`, return
        it, else, raise :exc:`MaskedObjectException` if the requested object is
        masked"""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)

                with t.inner():
                    result = method(*args, **kwargs)

                if result is None:
                    return None

                signature = inspect.signature(getattr(StorageInterface, method_name))
                bound_args = signature.bind(self, *args, **kwargs)
                self._raise_if_masked_swhids(
                    self._get_swhids_in_args(method_name, bound_args.arguments)
                )

                return result

        return newf

    RANDOM_ATTEMPTS = 5

    def _getter_random(self, method_name: str):
        """Handles methods returning a random object. Try
        :const:`RANDOM_ATTEMPTS` times for a non-masked object, and return it,
        else return :const:`None`."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)
                for _ in range(self.RANDOM_ATTEMPTS):
                    with t.inner():
                        result = method(*args, **kwargs)

                    if result is None:
                        return None

                    if not self._masked_result(method_name, result):
                        return result

        return newf

    def _getter_optional(self, method_name: str):
        """Handles methods returning an optional object: if the return value is
        :const:`None`, return it, else, raise a :exc:`MaskedObjectException` if
        the return value should be masked."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)
                with t.inner():
                    result = method(*args, **kwargs)
                if result is None:
                    return None

                self._raise_if_masked_result(method_name, result)

                return result

        return newf

    def _raise_if_masked_result_in_list(
        self, method_name: str, results: Iterable[TResult]
    ) -> List[TResult]:
        """Raise a :exc:`MaskedObjectException` if any non-:const:`None` object
        in ``results`` is masked."""
        result_swhids = set()
        results = list(results)
        for result in results:
            if result is not None:
                result_swhids.update(self._get_swhids_in_result(method_name, result))

        if result_swhids:
            self._raise_if_masked_swhids(list(result_swhids))
        return results

    def _getter_list(
        self,
        method_name: str,
    ):
        """Handle methods returning a list (or a generator) of optional objects,
        raising :exc:`MaskedObjectException` for all the masked objects in the
        batch."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)

                with t.inner():
                    results = list(method(*args, **kwargs))

                self._raise_if_masked_result_in_list(method_name, results)
                return results

        return newf

    def _getter_pagedresult(self, method_name: str) -> Callable[..., PagedResult]:
        """Handle methods returning a :cls:`PagedResult`, raising
        :exc:`MaskedObjectException` if some objects in the returned page are
        masked."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs) -> PagedResult:
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)
                with t.inner():
                    results = method(*args, **kwargs)

                self._raise_if_masked_result_in_list(method_name, results.results)

                return results

        return newf

    # Patching proxy feature set

    TRevision = TypeVar("TRevision", Revision, Optional[Revision])

    def _apply_revision_display_names(
        self, revisions: List[TRevision]
    ) -> List[TRevision]:
        emails = set()
        for rev in revisions:
            if (
                rev is not None
                and rev.author is not None
                and rev.author.email  # ignore None or empty email addresses
            ):
                emails.add(rev.author.email)
            if (
                rev is not None
                and rev.committer is not None
                and rev.committer.email  # ignore None or empty email addresses
            ):
                emails.add(rev.committer.email)

        with self._masking_query() as q:
            display_names = q.display_name(list(emails))

        # Short path for the common case
        if not display_names:
            return revisions

        persons: Dict[Optional[bytes], Person] = {
            email: Person.from_fullname(display_name)
            for (email, display_name) in display_names.items()
        }

        return [
            (
                None
                if revision is None
                else attr.evolve(
                    revision,
                    author=(
                        revision.author
                        if revision.author is None
                        else persons.get(revision.author.email, revision.author)
                    ),
                    committer=(
                        revision.committer
                        if revision.committer is None
                        else persons.get(revision.committer.email, revision.committer)
                    ),
                )
            )
            for revision in revisions
        ]

    TRelease = TypeVar("TRelease", Release, Optional[Release])

    def _apply_release_display_names(self, releases: List[TRelease]) -> List[TRelease]:
        emails = set()
        for rel in releases:
            if (
                rel is not None
                and rel.author is not None
                and rel.author.email  # ignore None or empty email addresses
            ):
                emails.add(rel.author.email)

        with self._masking_query() as q:
            display_names = q.display_name(list(emails))
        # Short path for the common case
        if not display_names:
            return releases

        persons: Dict[Optional[bytes], Person] = {
            email: Person.from_fullname(display_name)
            for (email, display_name) in display_names.items()
        }

        return [
            (
                None
                if release is None
                else attr.evolve(
                    release,
                    author=(
                        release.author
                        if release.author is None
                        else persons.get(release.author.email, release.author)
                    ),
                )
            )
            for release in releases
        ]

    def revision_get(
        self, revision_ids: List[Sha1Git], ignore_displayname: bool = False
    ) -> List[Optional[Revision]]:
        revisions = self.storage.revision_get(revision_ids)
        self._raise_if_masked_result_in_list("revision_get", revisions)
        return self._apply_revision_display_names(revisions)

    def revision_log(
        self,
        revisions: List[Sha1Git],
        ignore_displayname: bool = False,
        limit: Optional[int] = None,
    ) -> Iterable[Optional[Dict[str, Any]]]:
        revision_batches = grouper(
            self.storage.revision_log(revisions, limit=limit), BATCH_SIZE
        )
        yield from map(
            Revision.to_dict,
            itertools.chain.from_iterable(
                self._apply_revision_display_names(
                    self._raise_if_masked_result_in_list(
                        "revision_log", list(map(Revision.from_dict, revision_batch))
                    )
                )
                for revision_batch in revision_batches
            ),
        )

    def revision_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Revision]:
        page: PagedResult[Revision] = self.storage.revision_get_partition(
            partition_id, nb_partitions, page_token, limit
        )
        return PagedResult(
            results=self._apply_revision_display_names(
                self._raise_if_masked_result_in_list(
                    "revision_get_parition", page.results
                )
            ),
            next_page_token=page.next_page_token,
        )

    def release_get(
        self, releases: List[Sha1Git], ignore_displayname: bool = False
    ) -> List[Optional[Release]]:
        return self._apply_release_display_names(
            self._raise_if_masked_result_in_list(
                "release_get", self.storage.release_get(releases)
            )
        )

    def release_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Release]:
        page = self.storage.release_get_partition(
            partition_id, nb_partitions, page_token, limit
        )
        return PagedResult(
            results=self._apply_release_display_names(
                self._raise_if_masked_result_in_list(
                    "release_get_partition", page.results
                )
            ),
            next_page_token=page.next_page_token,
        )



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import math
from typing import Any, Dict, Optional, Tuple
import warnings

from swh.core.utils import encode_with_unescape
from swh.model.model import (
    ExtID,
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    ObjectType,
    Origin,
    Person,
    RawExtrinsicMetadata,
    Release,
    Revision,
    RevisionType,
    Timestamp,
    TimestampWithTimezone,
)
from swh.model.swhids import CoreSWHID
from swh.model.swhids import ExtendedObjectType as SwhidExtendedObjectType
from swh.model.swhids import ExtendedSWHID
from swh.model.swhids import ObjectType as SwhidObjectType
from swh.storage.interface import ObjectReference

from ..utils import map_optional

DEFAULT_AUTHOR = {
    "fullname": None,
    "name": None,
    "email": None,
}

DEFAULT_DATE = {
    "timestamp": None,
    "offset": 0,
    "neg_utc_offset": None,
    "offset_bytes": None,
}


def author_to_db(author: Optional[Person]) -> Dict[str, Any]:
    """Convert a swh-model author to its DB representation.

    Args:
        author: a :mod:`swh.model` compatible author

    Returns:
        dict: a dictionary with three keys: author, fullname and email

    """
    if author is None:
        return DEFAULT_AUTHOR

    return author.to_dict()


def db_to_author(
    fullname: Optional[bytes], name: Optional[bytes], email: Optional[bytes]
) -> Optional[Person]:
    """Convert the DB representation of an author to a swh-model author.

    Args:
        fullname (bytes): the author's fullname
        name (bytes): the author's name
        email (bytes): the author's email

    Returns:
        a Person object, or None if 'fullname' is None.
    """
    if fullname is None:
        return None

    if name is None and email is None:
        # The fullname hasn't been parsed, try that again
        return Person.from_fullname(fullname)

    return Person(
        fullname=fullname,
        name=name,
        email=email,
    )


def db_to_git_headers(db_git_headers):
    ret = []
    for key, value in db_git_headers:
        ret.append([key.encode("utf-8"), encode_with_unescape(value)])
    return ret


def db_to_date(
    date: Optional[datetime.datetime],
    offset_bytes: bytes,
) -> Optional[TimestampWithTimezone]:
    """Convert the DB representation of a date to a swh-model compatible date.

    Args:
        date: a date pulled out of the database
        offset_bytes: a byte representation of the latter two, usually as "+HHMM"
          or "-HHMM"

    Returns:
        a TimestampWithTimezone, or None if the date is None.

    """

    if date is None:
        return None

    return TimestampWithTimezone(
        timestamp=Timestamp(
            # we use floor() instead of int() to round down, because of negative dates
            seconds=math.floor(date.timestamp()),
            microseconds=date.microsecond,
        ),
        offset_bytes=offset_bytes,
    )


def date_to_db(ts_with_tz: Optional[TimestampWithTimezone]) -> Dict[str, Any]:
    """Convert a swh-model date_offset to its DB representation.

    Args:
        ts_with_tz: a TimestampWithTimezone object

    Returns:
        dict: a dictionary with these keys:

            - timestamp: a date in ISO format
            - offset_bytes: a byte representation of the latter two, usually as "+HHMM"
              or "-HHMM"

    """

    if ts_with_tz is None:
        return DEFAULT_DATE

    ts = ts_with_tz.timestamp

    timestamp = datetime.datetime.fromtimestamp(ts.seconds, datetime.timezone.utc)
    timestamp = timestamp.replace(microsecond=ts.microseconds)

    return {
        # PostgreSQL supports isoformatted timestamps
        "timestamp": timestamp.isoformat(),
        "offset": ts_with_tz.offset_minutes(),
        "neg_utc_offset": ts_with_tz.offset_minutes() == 0
        and ts_with_tz.offset_bytes.startswith(b"-"),
        "offset_bytes": ts_with_tz.offset_bytes,
    }


def revision_to_db(revision: Revision) -> Dict[str, Any]:
    """Convert a swh-model revision to its database representation."""

    author = author_to_db(revision.author)
    date = date_to_db(revision.date)
    committer = author_to_db(revision.committer)
    committer_date = date_to_db(revision.committer_date)

    return {
        "id": revision.id,
        "author_fullname": author["fullname"],
        "author_name": author["name"],
        "author_email": author["email"],
        "date": date["timestamp"],
        "date_offset": date["offset"],
        "date_neg_utc_offset": date["neg_utc_offset"],
        "date_offset_bytes": date["offset_bytes"],
        "committer_fullname": committer["fullname"],
        "committer_name": committer["name"],
        "committer_email": committer["email"],
        "committer_date": committer_date["timestamp"],
        "committer_date_offset": committer_date["offset"],
        "committer_date_neg_utc_offset": committer_date["neg_utc_offset"],
        "committer_date_offset_bytes": committer_date["offset_bytes"],
        "type": revision.type.value,
        "directory": revision.directory,
        "message": revision.message,
        "metadata": None if revision.metadata is None else dict(revision.metadata),
        "synthetic": revision.synthetic,
        "extra_headers": revision.extra_headers,
        "raw_manifest": revision.raw_manifest,
        "parents": [
            {
                "id": revision.id,
                "parent_id": parent,
                "parent_rank": i,
            }
            for i, parent in enumerate(revision.parents)
        ],
    }


def db_to_optional_revision(db_revision: Dict[str, Any]) -> Optional[Revision]:
    """Convert a database representation of a revision to its swh-model
    representation.
    This is similar to :func:`db_to_revision`, but this returns :const:`None`
    instead of crashing if the revision does not exist (returned by an outer join).
    """
    if db_revision["type"] is None:
        assert all(
            v is None for (k, v) in db_revision.items() if k not in ("id", "parents")
        )
        return None

    return db_to_revision(db_revision)


def db_to_revision(db_revision: Dict[str, Any]) -> Revision:
    """Convert a database representation of a revision to its swh-model
    representation. Raises ValueError if required values are None."""
    author = db_to_author(
        db_revision["author_fullname"],
        db_revision["author_name"],
        db_revision["author_email"],
    )
    date = db_to_date(
        db_revision["date"],
        db_revision["date_offset_bytes"],
    )

    committer = db_to_author(
        db_revision["committer_fullname"],
        db_revision["committer_name"],
        db_revision["committer_email"],
    )
    committer_date = db_to_date(
        db_revision["committer_date"],
        db_revision["committer_date_offset_bytes"],
    )

    assert (author is None) == (
        db_revision["author_fullname"] is None
    ), "author is unexpectedly None"
    assert (committer is None) == (
        db_revision["committer_fullname"] is None
    ), "committer is unexpectedly None"

    parents = []
    if "parents" in db_revision:
        for parent in db_revision["parents"]:
            if parent:
                parents.append(parent)

    metadata = db_revision["metadata"]
    extra_headers = db_revision["extra_headers"]
    if not extra_headers:
        if metadata and "extra_headers" in metadata:
            extra_headers = db_to_git_headers(metadata.pop("extra_headers"))
        else:
            # For older versions of the database that were not migrated to schema v161
            extra_headers = ()

    return Revision(
        id=db_revision["id"],
        author=author,
        date=date,
        committer=committer,
        committer_date=committer_date,
        type=RevisionType(db_revision["type"]),
        directory=db_revision["directory"],
        message=db_revision["message"],
        metadata=metadata,
        synthetic=db_revision["synthetic"],
        extra_headers=extra_headers,
        parents=tuple(parents),
        raw_manifest=db_revision["raw_manifest"],
    )


def release_to_db(release: Release) -> Dict[str, Any]:
    """Convert a swh-model release to its database representation."""
    author = author_to_db(release.author)
    date = date_to_db(release.date)

    return {
        "id": release.id,
        "author_fullname": author["fullname"],
        "author_name": author["name"],
        "author_email": author["email"],
        "date": date["timestamp"],
        "date_offset": date["offset"],
        "date_neg_utc_offset": date["neg_utc_offset"],
        "date_offset_bytes": date["offset_bytes"],
        "name": release.name,
        "target": release.target,
        "target_type": release.target_type.value,
        "comment": release.message,
        "synthetic": release.synthetic,
        "raw_manifest": release.raw_manifest,
    }


def db_to_optional_release(db_release: Dict[str, Any]) -> Optional[Release]:
    """Convert a database representation of a release to its swh-model
    representation.
    This is similar to :func:`db_to_release`, but this returns :const:`None`
    instead of crashing if the revision does not exist (returned by an outer join).
    """
    if db_release["target_type"] is None:
        assert all(v is None for (k, v) in db_release.items() if k != "id")
        return None

    return db_to_release(db_release)


def db_to_release(db_release: Dict[str, Any]) -> Release:
    """Convert a database representation of a release to its swh-model
    representation. Raises ValueError if required values are None.
    """
    author = db_to_author(
        db_release["author_fullname"],
        db_release["author_name"],
        db_release["author_email"],
    )
    date = db_to_date(
        db_release["date"],
        db_release["date_offset_bytes"],
    )

    return Release(
        author=author,
        date=date,
        id=db_release["id"],
        name=db_release["name"],
        message=db_release["comment"],
        synthetic=db_release["synthetic"],
        target=db_release["target"],
        target_type=ObjectType(db_release["target_type"]),
        raw_manifest=db_release["raw_manifest"],
    )


def db_to_raw_extrinsic_metadata(row) -> RawExtrinsicMetadata:
    target = row["raw_extrinsic_metadata.target"]
    if not target.startswith("swh:1:"):
        warnings.warn(
            "Fetching raw_extrinsic_metadata row with URL target", DeprecationWarning
        )
        target = str(Origin(url=target).swhid())

    return RawExtrinsicMetadata(
        target=ExtendedSWHID.from_string(target),
        authority=MetadataAuthority(
            type=MetadataAuthorityType(row["metadata_authority.type"]),
            url=row["metadata_authority.url"],
        ),
        fetcher=MetadataFetcher(
            name=row["metadata_fetcher.name"],
            version=row["metadata_fetcher.version"],
        ),
        discovery_date=row["discovery_date"],
        format=row["format"],
        metadata=row["raw_extrinsic_metadata.metadata"],
        origin=row["origin"],
        visit=row["visit"],
        snapshot=map_optional(CoreSWHID.from_string, row["snapshot"]),
        release=map_optional(CoreSWHID.from_string, row["release"]),
        revision=map_optional(CoreSWHID.from_string, row["revision"]),
        path=row["path"],
        directory=map_optional(CoreSWHID.from_string, row["directory"]),
    )


def db_to_extid(row) -> ExtID:
    return ExtID(
        extid=row["extid"],
        extid_type=row["extid_type"],
        extid_version=row.get("extid_version", 0),
        target=CoreSWHID(
            object_id=row["target"],
            object_type=SwhidObjectType[row["target_type"].upper()],
        ),
    )


def object_reference_to_db(
    reference: ObjectReference,
) -> Tuple[str, bytes, str, bytes]:
    return (
        reference.source.object_type.name.lower(),
        reference.source.object_id,
        reference.target.object_type.name.lower(),
        reference.target.object_id,
    )


def db_to_object_reference_source(row) -> ExtendedSWHID:
    return ExtendedSWHID(
        object_id=row["source"],
        object_type=SwhidExtendedObjectType[row["source_type"].upper()],
    )



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Iterable, List, Optional, Tuple

from swh.core.api.classes import stream_results_optional
from swh.model.model import Directory, DirectoryEntry, Sha1Git
from swh.storage.interface import StorageInterface


def directory_get(
    storage: StorageInterface, directory_id: Sha1Git
) -> Optional[Directory]:
    """Get all the entries for a given directory

    Args:
        storage: the storage instance
        directory_id: the directory's identifier
    Returns:
        The directory if it could be properly put back together.
    """
    entries: Optional[Iterable[DirectoryEntry]] = stream_results_optional(
        storage.directory_get_entries,
        directory_id=directory_id,
    )
    if entries is None:
        return None
    return Directory(
        id=directory_id,
        entries=tuple(entries),
        raw_manifest=storage.directory_get_raw_manifest([directory_id])[directory_id],
    )


def directory_get_many(
    storage: StorageInterface, directory_ids: List[Sha1Git]
) -> Iterable[Optional[Directory]]:
    """Same as :func:`directory_get`, but fetches directories slightly more
    effectively by batching requests to ``directory_get_raw_manifest``.

    Args:
        storage: the storage instance
        directory_ids: the directories' identifiers
    Yields:
        The directories which could be properly put back together
    """
    raw_manifests = storage.directory_get_raw_manifest(directory_ids)
    for directory_id in directory_ids:
        if directory_id not in raw_manifests:
            yield None
        else:
            entries = stream_results_optional(
                storage.directory_get_entries,
                directory_id=directory_id,
            )
            assert entries, f"Directory {directory_id.hex()} stopped existing"
            yield Directory(
                id=directory_id,
                entries=tuple(entries),
                raw_manifest=raw_manifests[directory_id],
            )


def directory_get_many_with_possibly_duplicated_entries(
    storage: StorageInterface, directory_ids: List[Sha1Git]
) -> Iterable[Optional[Tuple[bool, Directory]]]:
    """Same as :func:`directory_get_many`, but does not error on directories whose
    entries may contain duplicated names.
    See :meth:`swh.model.model.Directory.from_possibly_duplicated_entries`.

    Args:
        storage: the storage instance
        directory_ids: the directories' identifiers
    Yields:
        ``(is_corrupt, directory)`` where ``is_corrupt`` is True iff some
        entry names were indeed duplicated
    """
    raw_manifests = storage.directory_get_raw_manifest(directory_ids)
    for directory_id in directory_ids:
        if directory_id not in raw_manifests:
            yield None
        else:
            entries = stream_results_optional(
                storage.directory_get_entries,
                directory_id=directory_id,
            )
            assert entries, f"Directory {directory_id.hex()} stopped existing"
            yield Directory.from_possibly_duplicated_entries(
                id=directory_id,
                entries=tuple(entries),
                raw_manifest=raw_manifests[directory_id],
            )



# Copyright (C) 2019-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dataclasses
import datetime
import functools
import logging
import re
from typing import Any, Container, Dict, Optional, Tuple, cast

import attr
import pytest

from swh.journal.client import JournalClient
from swh.journal.serializers import kafka_to_value, key_to_kafka, value_to_kafka
from swh.journal.writer import JournalWriterInterface
from swh.model.hashutil import MultiHash, hash_to_bytes, hash_to_hex
from swh.model.model import ModelObjectType, Revision, RevisionType
from swh.model.tests.swh_model_data import (
    COMMITTERS,
    DATES,
    DUPLICATE_CONTENTS,
    REVISIONS,
)
from swh.model.tests.swh_model_data import TEST_OBJECTS as _TEST_OBJECTS
from swh.storage import get_storage
from swh.storage.cassandra.model import ContentRow, SkippedContentRow
from swh.storage.exc import StorageArgumentException
from swh.storage.in_memory import InMemoryStorage
from swh.storage.interface import StorageInterface
from swh.storage.replay import ModelObjectDeserializer, process_replay_objects

UTC = datetime.timezone.utc

TEST_OBJECTS = _TEST_OBJECTS.copy()
# add a revision with metadata to check this later is dropped while being replayed
TEST_OBJECTS[ModelObjectType.REVISION] = list(
    _TEST_OBJECTS[ModelObjectType.REVISION]
) + [
    Revision(
        id=hash_to_bytes("51d9d94ab08d3f75512e3a9fd15132e0a7ca7928"),
        message=b"hello again",
        date=DATES[1],
        committer=COMMITTERS[1],
        author=COMMITTERS[0],
        committer_date=DATES[0],
        type=RevisionType.GIT,
        directory=b"\x03" * 20,
        synthetic=False,
        metadata={"something": "interesting"},
        parents=(REVISIONS[0].id,),
    ),
]
WRONG_ID_REG = re.compile(
    "Object has id [0-9a-f]{40}, but it should be [0-9a-f]{40}: .*"
)


def nullify_ctime(obj: Any) -> Any:
    if isinstance(obj, (ContentRow, SkippedContentRow)):
        return dataclasses.replace(obj, ctime=None)
    else:
        return obj


@pytest.fixture()
def replayer_storage_and_client(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server: str
):
    journal_writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "auto_flush": False,
    }
    storage_config: Dict[str, Any] = {
        "cls": "memory",
        "journal_writer": journal_writer_config,
    }
    storage = get_storage(**storage_config)
    deserializer = ModelObjectDeserializer()
    replayer = JournalClient(
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_on_eof=True,
        value_deserializer=deserializer.convert,
    )

    yield storage, replayer


@pytest.mark.parametrize(
    "buffered",
    [pytest.param(True, id="buffered"), pytest.param(False, id="unbuffered")],
)
def test_storage_replayer(replayer_storage_and_client, caplog, buffered):
    """Optimal replayer scenario.

    This:
    - writes objects to a source storage
    - replayer consumes objects from the topic and replays them
    - a destination storage is filled from this

    In the end, both storages should have the same content.
    """
    src, replayer = replayer_storage_and_client

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)
    src.journal_writer.journal.flush()

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    if buffered:
        proxied_dst = get_storage("buffer", storage={"cls": "memory"})
        dst = proxied_dst.storage
    else:
        proxied_dst = dst = get_storage(cls="memory")

    # Fill the destination storage from Kafka
    worker_fn = functools.partial(process_replay_objects, storage=proxied_dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    assert isinstance(src, InMemoryStorage)  # needed to help mypy
    assert isinstance(dst, InMemoryStorage)
    check_replayed(src, dst)

    collision = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if "Colliding contents:" in logtext:
            collision += 1

    assert collision == 0, "No collision should be detected"


def test_storage_replay_with_collision(
    replayer_storage_and_client: Tuple[StorageInterface, JournalClient], caplog
):
    """Another replayer scenario with collisions.

    This:
    - writes objects to the topic, including colliding contents
    - replayer consumes objects from the topic and replay them
    - This drops the colliding contents from the replay when detected

    """
    src, replayer = replayer_storage_and_client
    journal: JournalWriterInterface = src.journal_writer.journal  # type: ignore
    assert journal is not None

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)
    journal.flush()

    # Create collision in input data
    # These should not be written in the destination
    producer = journal.producer  # type: ignore
    prefix = journal._prefix  # type: ignore
    for content in DUPLICATE_CONTENTS:
        topic = f"{prefix}.content"
        key = content.sha1
        now = datetime.datetime.now(tz=UTC)
        content = attr.evolve(content, ctime=now)
        producer.produce(
            topic=topic,
            key=key_to_kafka(key),
            value=value_to_kafka(content.to_dict()),
        )
        nb_sent += 1

    producer.flush()

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check the logs for the collision being properly detected
    nb_collisions = 0
    actual_collision: Dict
    for record in caplog.records:
        logtext = record.getMessage()
        if "Collision detected:" in logtext:
            nb_collisions += 1
            actual_collision = record.args["collision"]

    assert nb_collisions == 1, "1 collision should be detected"

    algo = "sha1"
    assert actual_collision["algo"] == algo
    expected_colliding_hash = hash_to_hex(DUPLICATE_CONTENTS[0].get_hash(algo))
    assert actual_collision["hash"] == expected_colliding_hash

    actual_colliding_hashes = actual_collision["objects"]
    assert len(actual_colliding_hashes) == len(DUPLICATE_CONTENTS)
    for content in DUPLICATE_CONTENTS:
        expected_content_hashes = {
            k: hash_to_hex(v) for k, v in content.hashes().items()
        }
        assert expected_content_hashes in actual_colliding_hashes

    # all objects from the src should exists in the dst storage
    assert isinstance(src, InMemoryStorage)  # needed to help mypy
    assert isinstance(dst, InMemoryStorage)  # needed to help mypy
    check_replayed(src, dst, exclude=["contents"])
    # but the dst has one content more (one of the 2 colliding ones)
    assert (
        len(list(src._cql_runner._contents.iter_all()))
        == len(list(dst._cql_runner._contents.iter_all())) - 1
    )


def test_replay_skipped_content(replayer_storage_and_client):
    """Test the 'skipped_content' topic is properly replayed."""
    src, replayer = replayer_storage_and_client
    _check_replay_skipped_content(src, replayer, "skipped_content")


# utility functions


def check_replayed(
    src: InMemoryStorage,
    dst: InMemoryStorage,
    exclude: Optional[Container] = None,
    expected_anonymized=False,
):
    """Simple utility function to compare the content of 2 in_memory storages"""

    def fix_expected(attr, row):
        if expected_anonymized:
            if attr == "releases":
                row = dataclasses.replace(
                    row, author=row.author and row.author.anonymize()
                )
            elif attr == "revisions":
                row = dataclasses.replace(
                    row,
                    author=row.author.anonymize(),
                    committer=row.committer.anonymize(),
                )
        if attr == "revisions":
            # the replayer should now drop the metadata attribute; see
            # swh/storgae/replay.py:_insert_objects()
            row.metadata = "null"

        return row

    for attr_ in (
        "contents",
        "skipped_contents",
        "directories",
        "extid",
        "revisions",
        "releases",
        "snapshots",
        "origins",
        "origin_visits",
        "origin_visit_statuses",
        "raw_extrinsic_metadata",
    ):
        if exclude and attr_ in exclude:
            continue
        expected_objects = [
            (id, nullify_ctime(fix_expected(attr_, obj)))
            for id, obj in sorted(getattr(src._cql_runner, f"_{attr_}").iter_all())
        ]
        got_objects = [
            (id, nullify_ctime(obj))
            for id, obj in sorted(getattr(dst._cql_runner, f"_{attr_}").iter_all())
        ]
        assert got_objects == expected_objects, f"Mismatch object list for {attr_}"


def _check_replay_skipped_content(storage, replayer, topic):
    skipped_contents = _gen_skipped_contents(100)
    nb_sent = len(skipped_contents)
    producer = storage.journal_writer.journal.producer
    prefix = storage.journal_writer.journal._prefix

    for i, obj in enumerate(skipped_contents):
        producer.produce(
            topic=f"{prefix}.{topic}",
            key=key_to_kafka({"sha1": obj["sha1"]}),
            value=value_to_kafka(obj),
        )
    producer.flush()

    dst_storage = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst_storage)
    nb_inserted = replayer.process(worker_fn)

    assert nb_sent == nb_inserted
    for content in skipped_contents:
        assert not storage.content_find({"sha1": content["sha1"]})

    # no skipped_content_find API endpoint, so use this instead
    assert not list(dst_storage.skipped_content_missing(skipped_contents))


def _updated(d1, d2):
    d1.update(d2)
    d1.pop("data", None)
    return d1


def _gen_skipped_contents(n=10):
    # we do not use the hypothesis strategy here because this does not play well with
    # pytest fixtures, and it makes test execution very slow
    now = datetime.datetime.now(tz=UTC)
    return [
        _updated(
            MultiHash.from_data(data=f"foo{i}".encode()).digest(),
            {
                "status": "absent",
                "reason": "why not",
                "origin": f"https://somewhere/{i}",
                "ctime": now,
                "length": len(f"foo{i}"),
            },
        )
        for i in range(n)
    ]


@pytest.mark.parametrize("privileged", [True, False])
def test_storage_replay_anonymized(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
    privileged: bool,
):
    """Optimal replayer scenario.

    This:
    - writes objects to the topic
    - replayer consumes objects from the topic and replay them

    This tests the behavior with both a privileged and non-privileged replayer
    """
    writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "anonymize": True,
        "auto_flush": False,
    }
    src_config: Dict[str, Any] = {"cls": "memory", "journal_writer": writer_config}

    storage = get_storage(**src_config)

    # Fill the src storage
    nb_sent = 0
    for obj_type, objs in TEST_OBJECTS.items():
        if obj_type in (
            ModelObjectType.ORIGIN_VISIT,
            ModelObjectType.ORIGIN_VISIT_STATUS,
        ):
            # these are unrelated with what we want to test here
            continue
        method = getattr(storage, f"{obj_type}_add")
        method(objs)
        nb_sent += len(objs)
    journal: JournalWriterInterface = storage.journal_writer.journal  # type:ignore
    journal.flush()

    # Fill a destination storage from Kafka, potentially using privileged topics
    dst_storage = get_storage(cls="memory")
    deserializer = ModelObjectDeserializer(
        validate=False
    )  # we cannot validate an anonymized replay
    replayer = JournalClient(
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_after_objects=nb_sent,
        privileged=privileged,
        value_deserializer=deserializer.convert,
    )
    worker_fn = functools.partial(process_replay_objects, storage=dst_storage)

    nb_inserted = replayer.process(worker_fn)
    replayer.consumer.commit()

    assert nb_sent == nb_inserted
    # Check the contents of the destination storage, and whether the anonymization was
    # properly used
    assert isinstance(storage, InMemoryStorage)  # needed to help mypy
    assert isinstance(dst_storage, InMemoryStorage)
    check_replayed(storage, dst_storage, expected_anonymized=not privileged)


def test_storage_replayer_with_validation_ok(
    replayer_storage_and_client, caplog, redisdb
):
    """Optimal replayer scenario

    with validation activated and reporter set to a redis db.

    - writes objects to a source storage
    - replayer consumes objects from the topic and replays them
    - a destination storage is filled from this
    - nothing has been reported in the redis db
    - both storages should have the same content
    """
    src, replayer = replayer_storage_and_client
    replayer.deserializer = ModelObjectDeserializer(validate=True, reporter=redisdb.set)

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)
    src.journal_writer.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check we do not have invalid objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            invalid += 1
    assert invalid == 0, "Invalid objects should not be detected"
    assert not redisdb.keys()
    # so the dst should be the same as src storage
    check_replayed(cast(InMemoryStorage, src), cast(InMemoryStorage, dst))


def test_storage_replayer_with_validation_nok(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer scenario with invalid objects

    with validation and reporter set to a redis db.

    - writes objects to a source storage
    - replayer consumes objects from the topic and replays them
    - the destination storage is filled with only valid objects
    - the redis db contains the invalid (raw kafka mesg) objects
    """
    src, replayer = replayer_storage_and_client
    replayer.value_deserializer = ModelObjectDeserializer(
        validate=True, reporter=redisdb.set
    ).convert

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)

    # insert invalid objects
    for object_type in (
        ModelObjectType.REVISION,
        ModelObjectType.DIRECTORY,
        ModelObjectType.RELEASE,
        ModelObjectType.SNAPSHOT,
    ):
        method = getattr(src, f"{object_type}_add")
        method([attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)])
        nb_sent += 1
    # also add an object that won't even be possible to instantiate; this needs
    # to be done at low kafka level (since we cannot instantiate the invalid model
    # object...)
    # we use directory[1] because it actually have some entries
    dict_repr = {
        # copy each dir entry twice
        "entries": TEST_OBJECTS[ModelObjectType.DIRECTORY][1].to_dict()["entries"] * 2,
        "id": b"\x01" * 20,
    }
    topic = f"{src.journal_writer.journal._prefix}.directory"
    src.journal_writer.journal.send(topic, dict_repr["id"], dict_repr)
    nb_sent += 1

    src.journal_writer.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check we do have invalid objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            invalid += 1
    assert invalid == 4, "Invalid objects should be detected"
    assert set(redisdb.keys()) == {
        f"swh:1:{typ}:{'0'*40}".encode() for typ in ("rel", "rev", "snp", "dir")
    } | {b"directory:" + b"01" * 20}

    for key in redisdb.keys():
        # check the stored value looks right
        rawvalue = redisdb.get(key)
        value = kafka_to_value(rawvalue)
        assert isinstance(value, dict)
        assert "id" in value
        assert value["id"] in (b"\x00" * 20, b"\x01" * 20)

    # check that invalid objects did not reach the dst storage
    for attr_ in (
        "directories",
        "revisions",
        "releases",
        "snapshots",
    ):
        for id, obj in sorted(getattr(dst._cql_runner, f"_{attr_}").iter_all()):
            assert id not in (b"\x00" * 20, b"\x01" * 20)

    # check that valid objects did reach the dst storage
    # revisions
    expected = [
        attr.evolve(rev, metadata=None)
        for rev in TEST_OBJECTS[ModelObjectType.REVISION]
    ]
    result = dst.revision_get(
        [obj.id for obj in TEST_OBJECTS[ModelObjectType.REVISION]]
    )
    assert result == expected
    # releases
    expected = TEST_OBJECTS[ModelObjectType.RELEASE]
    result = dst.release_get([obj.id for obj in TEST_OBJECTS[ModelObjectType.RELEASE]])
    assert result == expected
    # snapshot
    # result from snapshot_get is paginated, so adapt the expected to be comparable
    expected = [
        {"next_branch": None, **obj.to_dict()}
        for obj in TEST_OBJECTS[ModelObjectType.SNAPSHOT]
    ]
    result = [
        dst.snapshot_get(obj.id) for obj in TEST_OBJECTS[ModelObjectType.SNAPSHOT]
    ]
    assert result == expected
    # directories
    for directory in TEST_OBJECTS[ModelObjectType.DIRECTORY]:
        assert set(dst.directory_get_entries(directory.id).results) == set(
            directory.entries
        )


def test_storage_replayer_with_validation_nok_with_exceptions(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer scenario with invalid objects, with exceptions

    with validation and reporter set to a redis db.

    - writes invalid objects to a source storage
    - replayer consumes objects from the topic and replays them (invalid
      objects being in the exception list)
    - the destination storage is filled with all objects
    - the redis db does not contain the invalid objects

    """
    src, replayer = replayer_storage_and_client

    caplog.set_level(logging.WARNING, "swh.journal.replay")

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)

    # insert invalid objects
    known_invalid = []
    all_objs = {}
    for object_type in (
        ModelObjectType.REVISION,
        ModelObjectType.DIRECTORY,
        ModelObjectType.RELEASE,
        ModelObjectType.SNAPSHOT,
    ):
        all_objs[object_type] = TEST_OBJECTS[object_type][:]
        method = getattr(src, f"{object_type}_add")
        invalid_obj = attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)
        method([invalid_obj])
        nb_sent += 1
        all_objs[object_type].append(invalid_obj)
        known_invalid.append(
            (object_type, b"\x00" * 20, TEST_OBJECTS[object_type][0].id)
        )

    replayer.value_deserializer = ModelObjectDeserializer(
        validate=True,
        reporter=redisdb.set,
        known_mismatched_hashes=tuple(known_invalid),
    ).convert
    src.journal_writer.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check we do have invalid but replicates objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            assert logtext.endswith("Known exception, replicating it anyway.")
            invalid += 1
    assert invalid == 4, "Replayed invalid objects should be detected"
    assert not set(redisdb.keys())

    # check that all objects did reach the dst storage
    # revisions
    expected = [attr.evolve(rev, metadata=None) for rev in all_objs["revision"]]
    result = dst.revision_get([obj.id for obj in all_objs["revision"]])
    assert result == expected
    # releases
    expected = all_objs["release"]
    result = dst.release_get([obj.id for obj in all_objs["release"]])
    assert result == expected
    # snapshot
    # result from snapshot_get is paginated, so adapt the expected to be comparable
    expected = [{"next_branch": None, **obj.to_dict()} for obj in all_objs["snapshot"]]
    result = [dst.snapshot_get(obj.id) for obj in all_objs["snapshot"]]
    assert result == expected
    # directories
    for directory in all_objs["directory"]:
        assert set(dst.directory_get_entries(directory.id).results) == set(
            directory.entries
        )


def test_storage_replayer_with_validation_nok_raises(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer scenario with invalid objects

    with raise_on_error set to True

    This:
    - writes both valid & invalid objects to a source storage
    - a StorageArgumentException should be raised while replayer consumes
      objects from the topic and replays them
    """
    src, replayer = replayer_storage_and_client
    replayer.value_deserializer = ModelObjectDeserializer(
        validate=True, reporter=redisdb.set, raise_on_error=True
    ).convert

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)

    # insert invalid objects
    for object_type in (
        ModelObjectType.REVISION,
        ModelObjectType.DIRECTORY,
        ModelObjectType.RELEASE,
        ModelObjectType.SNAPSHOT,
    ):
        method = getattr(src, f"{object_type}_add")
        method([attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)])
        nb_sent += 1

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    with pytest.raises(StorageArgumentException):
        replayer.process(worker_fn)

    # check we do have invalid objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            invalid += 1
    assert invalid == 1, "One invalid objects should be detected"
    assert len(redisdb.keys()) == 1
